/*
COPYRIGHT NOTICE
Copyright (c) 2005  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott;

import edu.ucla.loni.dataflow.translate.image.BufferedImageAccessor;
import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import java.awt.image.BufferedImage;

/**
 * Processor that returns properties of a Buffered Image.
 *
 * @version 26 May 2005
 */
public class AccessorToImage extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Extracts a Buffered Image from a Buffered " +
                               "Image Accessor.";

  /** Input Node name for the Buffered Image Accessor. */
  private static String ACCESSOR_NAME = "Accessor";

  /** Input Node description for the Buffered Image Accessor. */
  private static String ACCESSOR_DESC = "Accessor to extract from.";

  /** Output Node name for the extracted Buffered Image. */
  private static String IMAGE_NAME = "Image";

  /** Output Node description for the extracted Buffered Image. */
  private static String IMAGE_DESC = "Extracted image.";

  /**
   * Constructs an Accessor To Image with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public AccessorToImage(String name)
    {
      super(name, DESC);

      // Create the Input Node for the Buffered Image Accessor
      Class type = BufferedImageAccessor.class;
      InputNode inputNode = new InputNode(ACCESSOR_NAME, ACCESSOR_DESC, type);
      _addInputNode(inputNode);

      // Create the Output Node for the Buffered Image
      type = BufferedImage.class;
      OutputNode node = new OutputNode(IMAGE_NAME, IMAGE_DESC, type);
      _addOutputNode(node);
    }

  /**
   * Constructs an Accessor To Image using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public AccessorToImage(String name, ProcessorProperty[] properties)
    {
      this(name);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the Input value
	BufferedImageAccessor accessor =
	  (BufferedImageAccessor)_getInputValue(ACCESSOR_NAME, null);

	// If an Accessor is present, extract the Image
	BufferedImage image = null;
	if (accessor != null) { image = accessor.getRealBufferedImage(); }

	// Set the output value
	_setOutputValue(IMAGE_NAME, image);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }
}
