/*
  COPYRIGHT NOTICE
  Copyright (c) 2006  Scott C. Neu and Arthur W. Toga

  See README.license for license notices.
*/

package edu.ucla.loni.debabel.processors.scott;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import edu.ucla.loni.imageio.Utilities;
import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.image.DataBufferDouble;
import java.awt.image.DataBufferFloat;
import java.awt.image.DataBufferInt;
import java.awt.image.DataBufferShort;
import java.awt.image.DataBufferUShort;
import java.lang.reflect.Array;
import java.util.Vector;
import javax.imageio.ImageTypeSpecifier;

/**
 * Processor that maps the values of a Buffered Image into the values of
 * another Buffered Image (possibly of another data type).
 *
 * @version 14 August 2006
 */
public class BufferedImageMapper2 extends Processor
{
    /** Description of this Processor. */
    private static String DESC =
	"Linearly maps pixel values of the input image to output image values.";

    /** Input Node name for the image to map values from. */
    private static String ORIG_IMAGE_NAME = "Input Image";

    /** Input Node description for the image to map values from. */
    private static String ORIG_IMAGE_DESC =
	"Image to map pixel values from.";

    /** Input Node name for the input lower bound. */
    private static String INPUT_LOWER_BOUND_NAME = "Input Lower Bound";

    /** Input Node description for the input lower bound. */
    private static String INPUT_LOWER_BOUND_DESC =
	"Pixel value below which all values are mapped to the output lower " +
	"bound.";

    /** Input Node name for the input upper bound. */
    private static String INPUT_UPPER_BOUND_NAME = "Input Upper Bound";

    /** Input Node description for the input upper bound. */
    private static String INPUT_UPPER_BOUND_DESC =
	"Pixel value above which all values are mapped to the output upper " +
	"bound.";

    /** Output Node name for the new image. */
    private static String NEW_IMAGE_NAME = "Output Image";

    /** Output Node description for the new image. */
    private static String NEW_IMAGE_DESC = "Image containing mapped values.";

    /** Input Node name for the output lower bound. */
    private static String OUTPUT_LOWER_BOUND_NAME = "Output Lower Bound";

    /** Input Node description for the output lower bound. */
    private static String OUTPUT_LOWER_BOUND_DESC =
	"Smallest pixel value in the output image.";

    /** Input Node name for the output upper bound. */
    private static String OUTPUT_UPPER_BOUND_NAME = "Output Upper Bound";

    /** Input Node description for the output upper bound. */
    private static String OUTPUT_UPPER_BOUND_DESC =
	"Largest pixel value in the output image.";

    /** Property for the data type. */
    private static String PROP_TYPE_NAME = "type";

    /** Property description for the data type. */
    private static String PROP_TYPE_DESC = "Data type to convert to.";

    /**
     * Constructs a Buffered Image Mapper 2 with default Properties.
     *
     * @param name Name of the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public BufferedImageMapper2(String name)
    {
	this(name, _getDefaultProperties());
    }

    /**
     * Constructs a Buffered Image Mapper 2 using the specified Properties.
     *
     * @param name Name of the Processor.
     * @param properties Properties for the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public BufferedImageMapper2(String name, ProcessorProperty[] properties)
    {
	super(name, DESC);

	// Add the Properties
	_addProperties(_getDefaultProperties(), properties);

	// Create the image Input Node
	Class type = BufferedImage.class;
	InputNode inputNode = new InputNode(ORIG_IMAGE_NAME, ORIG_IMAGE_DESC,
					    type);
	_addInputNode(inputNode);

	// Create the input lower bound Node
	type = Double.class;
	inputNode = new InputNode(INPUT_LOWER_BOUND_NAME,
				  INPUT_LOWER_BOUND_DESC,
				  type);
	_addInputNode(inputNode);

	// Create the input upper bound Node
	type = Double.class;
	inputNode = new InputNode(INPUT_UPPER_BOUND_NAME,
				  INPUT_UPPER_BOUND_DESC,
				  type);
	_addInputNode(inputNode);

	// Create the output lower bound Node
	type = Double.class;
	inputNode = new InputNode(OUTPUT_LOWER_BOUND_NAME,
				  OUTPUT_LOWER_BOUND_DESC, type);
	_addInputNode(inputNode);

	// Create the output upper bound Node
	type = Double.class;
	inputNode = new InputNode(OUTPUT_UPPER_BOUND_NAME,
				  OUTPUT_UPPER_BOUND_DESC, type);
	_addInputNode(inputNode);

	// Create the Output Node
	type = BufferedImage.class;
	OutputNode node = new OutputNode(NEW_IMAGE_NAME, NEW_IMAGE_DESC, type);
	_addOutputNode(node);
    }

    /**
     * Generates values for the Output Nodes using the current values of the
     * Input Nodes.
     *
     * @throws ProcessingException If an error occurs during processing.
     */
    public void generateOutputValues() throws ProcessingException
    {
	try {

	    // Get the input Buffered Image
	    BufferedImage srcImage =
		(BufferedImage)_getInputValue(ORIG_IMAGE_NAME, null);
	    if (srcImage == null) { return; }

	    // Get the input data
	    Object srcData = _getData(srcImage);
	    if (srcData == null) { return; }

	    // Get the input lower bound
	    Double srcLower = (Double)_getInputValue(INPUT_LOWER_BOUND_NAME,
						     null);
	    if (srcLower == null) { return; }

	    // Get the input upper bound
	    Double srcUpper = (Double)_getInputValue(INPUT_UPPER_BOUND_NAME,
						     null);
	    if (srcUpper == null) { return; }

	    // Get the output lower bound
	    Double dstLower = (Double)_getInputValue(OUTPUT_LOWER_BOUND_NAME,
						     null);
	    if (dstLower == null) { return; }

	    // Get the output upper bound
	    Double dstUpper = (Double)_getInputValue(OUTPUT_UPPER_BOUND_NAME,
						     null);
	    if (dstUpper == null) { return; }

	    // Get the output data type
	    Class dstType = (Class)_getPropertyValue(PROP_TYPE_NAME);
	    int dstDataType = DataBuffer.TYPE_BYTE;
	    if (dstType == Short.class) {dstDataType = DataBuffer.TYPE_USHORT;}
	    if (dstType == Integer.class) { dstDataType = DataBuffer.TYPE_INT; }
	    if (dstType == Float.class) { dstDataType = DataBuffer.TYPE_FLOAT; }
	    if (dstType == Double.class) {dstDataType = DataBuffer.TYPE_DOUBLE;}
	
	    // Create the output Buffered Image
	    int w = srcImage.getWidth();
	    int h = srcImage.getHeight();
	    ImageTypeSpecifier spec = Utilities.getGrayImageType(dstDataType);
	    BufferedImage dstImage = spec.createBufferedImage(w, h);

	    // Get the output data
	    Object dstData = _getData(dstImage);

	    // Linearly map the input data into the output data
	    _linearMap(srcData, srcLower.doubleValue(), srcUpper.doubleValue(),
		       dstData, dstLower.doubleValue(), dstUpper.doubleValue());

	    // Return the output Buffered Image
	    _setOutputValue(NEW_IMAGE_NAME, dstImage);
	}

	// Unable to convert
	catch (Exception e) {
	    String msg = "Unable to generate output values for \"" + this +
		"\".";
	    throw new ProcessingException(msg, e);
	}
    }

    /**
     * Linearly maps the source data into the destination data.
     *
     * @param srcData Source data.
     * @param srcLower Lower bound to map source data from.
     * @param srcUpper Upper bound to map source data from.
     * @param dstData Destination data.
     * @param dstLower Lower bound to map destination data from.
     * @param dstUpper Upper bound to map destination data from.
     */
    private void _linearMap(Object srcData, double srcLower, double srcUpper,
			    Object dstData, double dstLower, double dstUpper)
    {
	// Infinite slope
	if (srcUpper == srcLower) { return; }

	// Compute the linear parameters
	double slope = (dstUpper - dstLower)/(srcUpper - srcLower);
	double offset = dstLower - srcLower*slope;

	// Map the data
	int length = Array.getLength(srcData);
	for (int i = 0; i < length; i++) {
	    double src = Array.getDouble(srcData, i);

	    // Assume unsigned bytes
	    if (srcData instanceof byte[]) {
		byte[] byteData = (byte[])srcData;
		src = byteData[i] & 0xff;
	    }

	    // Linearly map the data
	    double dst;
	    if (src <= srcLower) { dst = dstLower; }
	    else if (src >= srcUpper) { dst = dstUpper; }
	    else { dst = slope*src + offset; }

	    // Convert the mapped data to the appropriate type
	    if (dstData instanceof byte[]) {
		byte[] data = (byte[])dstData;
		data[i] = (byte)dst;
	    }
	    else if (dstData instanceof short[]) {
		short[] data = (short[])dstData;
		data[i] = (short)dst;
	    }
	    else if (dstData instanceof int[]) {
		int[] data = (int[])dstData;
		data[i] = (int)dst;
	    }
	    else if (dstData instanceof float[]) {
		float[] data = (float[])dstData;
		data[i] = (float)dst;
	    }
	    else {
		double[] data = (double[])dstData;
		data[i] = dst;
	    }
	}
    }

    /**
     * Gets the array of data in the specified Buffered Image.
     *
     * @param image Buffered Image.
     *
     * @return Array of data in the Buffered Image, or null if it cannot be
     *         found.
     */
    private Object _getData(BufferedImage image)
    {
	DataBuffer dataBuffer = image.getRaster().getDataBuffer();

	if (dataBuffer instanceof DataBufferByte) {
	    return ((DataBufferByte)dataBuffer).getData();
	}
	else if (dataBuffer instanceof DataBufferShort) {
	    return ((DataBufferShort)dataBuffer).getData();
	}
	else if (dataBuffer instanceof DataBufferUShort) {
	    return ((DataBufferUShort)dataBuffer).getData();
	}
	else if (dataBuffer instanceof DataBufferInt) {
	    return ((DataBufferInt)dataBuffer).getData();
	}
	else if (dataBuffer instanceof DataBufferFloat) {
	    return ((DataBufferFloat)dataBuffer).getData();
	}
	else if (dataBuffer instanceof DataBufferDouble) {
	    return ((DataBufferDouble)dataBuffer).getData();
	}

	// Unknown data type
	return null;
    }

    /**
     * Gets the default Properties.
     *
     * @return Default Processor Properties.
     */
    private static ProcessorProperty[] _getDefaultProperties()
    {
	ProcessorProperty[] properties = new ProcessorProperty[1];

	// Output data types
	Vector optionSet = new Vector();
	optionSet.add( Byte.class );
	optionSet.add( Short.class );
	optionSet.add( Integer.class );
	optionSet.add( Float.class );
	optionSet.add( Double.class );

	// Create a Set Restriction
	String name = "Output Image Data Types";
	String desc = "Data types that can be mapped to.";
	Restriction restriction = new SetRestriction(name, desc, optionSet);

	// Default output data type
	name = PROP_TYPE_NAME;
	desc = PROP_TYPE_DESC;
	Object value = Short.class;
	properties[0] = new ProcessorProperty(name, desc, value, restriction);

	return properties;
    }
}
