/*
  COPYRIGHT NOTICE
  Copyright (c) 2007  Scott C. Neu and Arthur W. Toga

  See README.license for license notices.
*/

package edu.ucla.loni.debabel.processors.scott;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import edu.ucla.loni.imageio.Utilities;
import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.image.DataBufferDouble;
import java.awt.image.DataBufferFloat;
import java.awt.image.DataBufferInt;
import java.awt.image.DataBufferShort;
import java.awt.image.DataBufferUShort;
import java.lang.reflect.Array;
import java.util.Vector;
import javax.imageio.ImageTypeSpecifier;

/**
 * Processor that rescales the pixel values of a Buffered Image.
 *
 * @version 10 April 2007
 */
public class BufferedImageRescaler extends Processor
{
    /** Description of this Processor. */
    private static String DESC = "Rescales a Buffered Image's pixel values.";

    /** Input Node name for the image to rescale. */
    private static String IMAGE_NAME = "Image";

    /** Input Node description for the image to rescale. */
    private static String IMAGE_DESC = "Image to rescale.";

    /** Input Node name for the rescale factor. */
    private static String FACTOR_NAME = "Rescale Factor";

    /** Input Node description for the rescale factor. */
    private static String FACTOR_DESC =
	"Factor to multiply each pixel value by.";

    /** Output Node name for the rescaled image. */
    private static String NEW_IMAGE_NAME = "Rescaled Image";

    /** Output Node description for the rescaled image. */
    private static String NEW_IMAGE_DESC = "Image that is rescaled.";

    /** Property for the data type. */
    private static String PROP_NAME = "type";

    /** Property description for the data type. */
    private static String PROP_DESC = "Data type of the rescaled image.";

    /**
     * Constructs a Buffered Image Rescaler with default Properties.
     *
     * @param name Name of the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public BufferedImageRescaler(String name)
    {
	this(name, _getDefaultProperties());
    }

    /**
     * Constructs a Buffered Image Rescaler using the specified Properties.
     *
     * @param name Name of the Processor.
     * @param properties Properties for the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public BufferedImageRescaler(String name, ProcessorProperty[] properties)
    {
	super(name, DESC);

	// Add the Properties
	_addProperties(_getDefaultProperties(), properties);

	// Create the Input Node for the image
	Class type = BufferedImage.class;
	InputNode inputNode = new InputNode(IMAGE_NAME, IMAGE_DESC, type);
	_addInputNode(inputNode);

	// Create the Input Node for the rescale factor
	type = Double.class;
	inputNode = new InputNode(FACTOR_NAME, FACTOR_DESC, type);
	_addInputNode(inputNode);

	// Create the Output Node
	type = BufferedImage.class;
	OutputNode node = new OutputNode(NEW_IMAGE_NAME, NEW_IMAGE_DESC, type);
	_addOutputNode(node);
    }

    /**
     * Generates values for the Output Nodes using the current values of the
     * Input Nodes.
     *
     * @throws ProcessingException If an error occurs during processing.
     */
    public void generateOutputValues() throws ProcessingException
    {
	try {

	    // Get the original Buffered Image
	    Object value = _getInputValue(IMAGE_NAME, null);
	    if ( !(value instanceof BufferedImage) ) { return; }
	    BufferedImage image = (BufferedImage)value;

	    // Get the rescale factor
	    Double rescaleFactor = (Double)_getInputValue(FACTOR_NAME, null);
	    double rescale = 1;
	    if (rescaleFactor != null) {rescale = rescaleFactor.doubleValue();}

	    // Get the rescaled image type
	    int dataType = DataBuffer.TYPE_FLOAT;
	    Class type = (Class)_getPropertyValue(PROP_NAME);
	    if (type == Double.class) { dataType = DataBuffer.TYPE_DOUBLE; }

	    // Create the rescaled image
	    int w = image.getWidth();
	    int h = image.getHeight();
	    ImageTypeSpecifier spec = Utilities.getGrayImageType(dataType);
	    BufferedImage rescaledImage = spec.createBufferedImage(w, h);

	    // Apply the rescale factor
	    Object srcData = _getData(image);
	    Object dstData = _getData(rescaledImage);
	    _rescale(srcData, dstData, rescale);

	    // Return the output Buffered Image
	    _setOutputValue(NEW_IMAGE_NAME, rescaledImage);
	}

	// Unable to convert
	catch (Exception e) {
	    String msg = "Unable to generate output values for \"" + this +
		"\".";
	    throw new ProcessingException(msg, e);
	}
    }

    /**
     * Rescales the image data.
     *
     * @param srcData Source data.
     * @param dstData Destination data.
     * @param rescale Rescale factor applied to the source data.
     */
    private void _rescale(Object srcData, Object dstData, double rescale)
    {
	// Rescale the data
	int length = Array.getLength(srcData);
	for (int i = 0; i < length; i++) {
	    double src = Array.getDouble(srcData, i);

	    // Rescale the data
	    if (dstData instanceof float[]) {
		float[] data = (float[])dstData;
		data[i] = (float)(rescale*src);
	    }
	    else {
		double[] data = (double[])dstData;
		data[i] = rescale*src;
	    }
	}
    }

    /**
     * Gets the array of data in the specified Buffered Image.
     *
     * @param image Buffered Image.
     *
     * @return Array of data in the Buffered Image, or null if it cannot be
     *         found.
     */
    private Object _getData(BufferedImage image)
    {
	DataBuffer dataBuffer = image.getRaster().getDataBuffer();

	if (dataBuffer instanceof DataBufferByte) {
	    return ((DataBufferByte)dataBuffer).getData();
	}
	else if (dataBuffer instanceof DataBufferShort) {
	    return ((DataBufferShort)dataBuffer).getData();
	}
	else if (dataBuffer instanceof DataBufferUShort) {
	    return ((DataBufferUShort)dataBuffer).getData();
	}
	else if (dataBuffer instanceof DataBufferInt) {
	    return ((DataBufferInt)dataBuffer).getData();
	}
	else if (dataBuffer instanceof DataBufferFloat) {
	    return ((DataBufferFloat)dataBuffer).getData();
	}
	else if (dataBuffer instanceof DataBufferDouble) {
	    return ((DataBufferDouble)dataBuffer).getData();
	}

	// Unknown data type
	return null;
    }

    /**
     * Gets the default Properties.
     *
     * @return Default Processor Properties.
     */
    private static ProcessorProperty[] _getDefaultProperties()
    {
	ProcessorProperty[] properties = new ProcessorProperty[1];

	// Data Types
	Vector optionSet = new Vector();
	optionSet.add( Float.class );
	optionSet.add( Double.class );

	// Create a Set Restriction
	String name = "Allowed Data Types";
	String desc = "Data types that can be rescaled to.";
	Restriction restriction = new SetRestriction(name, desc, optionSet);

	// Default primitive type
	name = PROP_NAME;
	desc = PROP_DESC;
	Object value = Float.class;
	properties[0] = new ProcessorProperty(name, desc, value, restriction);

	return properties;
    }
}
