/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferUShort;
import java.awt.image.WritableRaster;

/**
 * Processor that combines a contour and image into another image.
 *
 * @version 12 May 2003
 */
public class ImageAndContourCombiner extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Combines a contour with an image.";

  /** Name of the Input Node for the image. */
  private static String IMAGE_NAME = "Image";

  /** Description of the Input Node for the image. */
  private static String IMAGE_DESC = "Image to combine.";

  /** Name of the Input Node for the contour. */
  private static String CONTOUR_NAME = "Contour";

  /** Description of the Input Node for the contour. */
  private static String CONTOUR_DESC = "Contour to combine.";

  /** Name of the Output Node for the combined image and contour. */
  private static String COMBINED_NAME = "Image + Contour";

  /** Description of the Output Node for the combined image and contour. */
  private static String COMBINED_DESC =
  "Image containing the input image and contour.";

  /**
   * Constructs an Image And Contour Combiner with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public ImageAndContourCombiner(String name)
    {
      super(name, DESC);

      // Create the Input Node that accepts the image
      Class type = BufferedImage.class;
      InputNode inputNode = new InputNode(IMAGE_NAME, IMAGE_DESC, type);
      _addInputNode(inputNode);

      // Create the Input Node that accepts the contour
      type = DataBuffer.class;
      inputNode = new InputNode(CONTOUR_NAME, CONTOUR_DESC, type);
      _addInputNode(inputNode);

      // Create the Output Node that produces the image + contour
      type = BufferedImage[].class;
      OutputNode imgNode = new OutputNode(COMBINED_NAME, COMBINED_DESC, type);
      _addOutputNode(imgNode);
    }

  /**
   * Constructs an Image And Contour Combiner using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public ImageAndContourCombiner(String name, ProcessorProperty[] properties)
    {
      this(name);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the image and contour
	BufferedImage image = (BufferedImage)_getInputValue(IMAGE_NAME, null);
	DataBuffer contour = (DataBuffer)_getInputValue(CONTOUR_NAME, null);

	// Produces the image + contour
	BufferedImage[] newImages = new BufferedImage[0];
	if (image != null && contour != null) {

	  // Create a new blank image
	  int width = image.getWidth();
	  int height = image.getHeight();
	  newImages = new BufferedImage[1];
	  newImages[0] = new BufferedImage(width, height,
					   BufferedImage.TYPE_USHORT_GRAY);

	  // Get the output shorts
	  WritableRaster newRaster = newImages[0].getRaster();
	  DataBufferUShort buff = (DataBufferUShort)newRaster.getDataBuffer();
	  short[] newShorts = buff.getData();

	  // Copy the input image data to the output shorts
	  for (int j = 0; j < height; j++) {
	    for (int i = 0; i < width; i++) {
	      int rgb = image.getRGB(i,j);

	      // ASSUME BINARY DATA
	      if (rgb == -1) { newShorts[j*height+i] = 128; }
	    }
	  }

	  // Copy the contour to the output shorts
	  for (int i = 0; i < contour.getSize(); i++) {
	    int x = contour.getElem(0, i);
	    int y = contour.getElem(1, i);
	    newShorts[y*height+x] = 255;
	  }
	}

	// Set the image + contour
	_setOutputValue(COMBINED_NAME, newImages);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }
}
