/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;

/**
 * Processor that combines multiple strings together.
 *
 * @version 12 May 2003
 */
public class MultiStringConcatenator extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Combines multiple strings together.";

  /** Base Input Node name for the ith String. */
  private static String BASE_STRING_NAME = "String #";

  /** Input Node description for the ith String. */
  private static String STRING_DESC = "String to concatenate.";

  /** Output Node name for the combined String. */
  private static String COMBINED_NAME = "Combined String.";

  /** Output Node description for the combined String. */
  private static String COMBINED_DESC = "All the Strings concatenated.";

  /** Property name for the number of Strings. */
  private static String PROP_NAME = "number";

  /** Property description for the number of Strings. */
  private static String PROP_DESC = "Number of strings to concatenate.";

  /**
   * Constructs a Multi String Concatenator with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public MultiStringConcatenator(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs an Multi String Concatenator using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public MultiStringConcatenator(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Get the number of Strings
      int number = ((Integer)_getPropertyValue(PROP_NAME)).intValue();

      // Create the Input Nodes for each String
      for (int i = 0; i < number; i++) {
	name = BASE_STRING_NAME + i;
	InputNode inputNode = new InputNode(name, STRING_DESC, String.class);
	_addInputNode(inputNode);
      }

      // Create the Output Node for the concatenated String
      OutputNode outputNode = new OutputNode(COMBINED_NAME, COMBINED_DESC,
					     String.class);
      _addOutputNode(outputNode);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Create a buffer for the concatenated String
	StringBuffer buffer = new StringBuffer();

	// Concatenate the input Strings
	int number = ((Integer)_getPropertyValue(PROP_NAME)).intValue();
	for (int i = 0; i < number; i++) {
	  String value = (String)_getInputValue(BASE_STRING_NAME + i, "");
	  buffer.append( value );
	}

	// Set the combined String
	_setOutputValue(COMBINED_NAME, buffer.toString());
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[1];

      // Default number of Strings
      String name = PROP_NAME;
      String desc = PROP_DESC;
      Object value = new Integer(2);
      Restriction restriction = Utilities.getPositiveRestriction();
      properties[0] = new ProcessorProperty(name, desc, value, restriction);

      return properties;
    }
}
