/*
  COPYRIGHT NOTICE
  Copyright (c) 2007  Scott C. Neu and Arthur W. Toga

  See README.license for license notices.
*/

package edu.ucla.loni.debabel.processors.scott.dicom;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.util.GregorianCalendar;
import java.util.Vector;

/**
 * Processor that parses a DICOM date string.
 *
 * @version 12 January 2007
 */
public class DicomDateParser extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Parses a DICOM date string.";

  /** Input Node name for the DICOM date string. */
  private static String DATE_NAME = "DICOM Date String";

  /** Input Node description for the DICOM date string. */
  private static String DATE_DESC = "String in the format YYYYMMDD.";

  /** Output Node name for the parsed year. */
  private static String YEAR_NAME = "Year";

  /** Output Node description for the parsed year. */
  private static String YEAR_DESC = "Year parsed from the date string.";

  /** Output Node name for the parsed month. */
  private static String MONTH_NAME = "Month";

  /** Output Node description for the parsed month. */
  private static String MONTH_DESC =
  "Month (1-12) parsed from the date string.";

  /** Output Node name for the parsed day. */
  private static String DAY_NAME = "Day";

  /** Output Node description for the parsed day. */
  private static String DAY_DESC = "Day (1-31) parsed from the date string.";

  /** Output Node name for the parsed time. */
  private static String TIME_NAME = "Time";

  /** Output Node description for the parsed time. */
  private static String TIME_DESC =
  "Number of milliseconds since January 1, 1970.";

  /** Property name for choosing the year/month/day or time. */
  private static String PROP_NAME = "Output Format";

  /** Property description for choosing the year/month/day or time. */
  private static String PROP_DESC = "Format for the parsed output.";

  /**
   * Constructs a DICOM Date Parser with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public DicomDateParser(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a DICOM Date Parser using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public DicomDateParser(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Get the output format
      String format = (String)_getPropertyValue(PROP_NAME);

      // Create the Input Node
      Class type = String.class;
      _addInputNode( new InputNode(DATE_NAME, DATE_DESC, type) );

      // Create the Output Nodes
      if ( format.equals("Year/Month/Day") ) {
	type = Integer.class;
	_addOutputNode( new OutputNode(YEAR_NAME, YEAR_DESC, type) );
	_addOutputNode( new OutputNode(MONTH_NAME, MONTH_DESC, type) );
	_addOutputNode( new OutputNode(DAY_NAME, DAY_DESC, type) );
      }
      else {
	type = Long.class;
	_addOutputNode( new OutputNode(TIME_NAME, TIME_DESC, type) );
      }
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the date string
	String date = (String)_getInputValue(DATE_NAME, "");

	// Remove spaces before and after
	date = date.trim();

	// Old versions of DICOM support YYYY.MM.DD -> remove all .'s
	date = date.replaceAll("[.]", "");
	if ( date.length() < 8 ) { return; }

	// Parse the year, month, and day
	int year = Integer.parseInt( date.substring(0, 4) );
	int month = Integer.parseInt( date.substring(4, 6) );
	int day = Integer.parseInt( date.substring(6, 8) );

	// Get the output format
	String format = (String)_getPropertyValue(PROP_NAME);

	// Set the Output Nodes
	if ( format.equals("Year/Month/Day") ) {
	  _setOutputValue(YEAR_NAME, new Integer(year));
	  _setOutputValue(MONTH_NAME, new Integer(month));
	  _setOutputValue(DAY_NAME, new Integer(day));
	}
	else {

	  // Convert to milliseconds
	  GregorianCalendar cal = new GregorianCalendar(year, month-1, day);
	  long time = cal.getTimeInMillis();
	  _setOutputValue(TIME_NAME, new Long(time));
	}
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[1];

      // Create the Restriction set
      Vector optionSet = new Vector();
      optionSet.add("Year/Month/Day");
      optionSet.add("Time");

      // Create the Restriction
      String name = PROP_NAME;
      String desc = PROP_DESC;
      SetRestriction r = new SetRestriction(name, desc, optionSet);

      // Return the Property
      properties[0] = new ProcessorProperty(name, desc, "Year/Month/Day", r);
      return properties;
    }
}
