/*
COPYRIGHT NOTICE
Copyright (c) 2004  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott.dicom;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import java.util.Date;
import java.util.GregorianCalendar;

/**
 * Processor that formats a datetime into DICOM date and time strings.
 *
 * @version 18 February 2004
 */
public class DicomDateTimeFormatter extends Processor
{
  /** Description of this Processor. */
  private static String DESC =
  "Formats date/time into DICOM date and time strings.";

  /** Input Node name for the date/time string. */
  private static String DATETIME_NAME = "Date/Time";

  /** Input Node description for the date/time string. */
  private static String DATETIME_DESC =
  "Number of milliseconds since January 1, 1970.";

  /** Output Node name for the DICOM date string. */
  private static String DATE_NAME = "DICOM Date String";

  /** Output Node description for the DICOM date string. */
  private static String DATE_DESC = "String in the format YYYYMMDD.";

  /** Output Node name for the DICOM time string. */
  private static String TIME_NAME = "DICOM Time String";

  /** Output Node description for the DICOM time string. */
  private static String TIME_DESC = "String in the format hhmmss.frac.";

  /**
   * Constructs a DICOM Date/Time Parser with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public DicomDateTimeFormatter(String name)
    {
      super(name, DESC);

      // Create the Input Node
      Class type = Long.class;
      _addInputNode( new InputNode(DATETIME_NAME, DATETIME_DESC, type) );

      // Create the Output Nodes
      type = String.class;
      _addOutputNode( new OutputNode(DATE_NAME, DATE_DESC, type) );
      _addOutputNode( new OutputNode(TIME_NAME, TIME_DESC, type) );
    }

  /**
   * Constructs a DICOM Date/Time Parser using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public DicomDateTimeFormatter(String name, ProcessorProperty[] properties)
    {
      this(name);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the date/time
	Long object = (Long)_getInputValue(DATETIME_NAME, null);
	if (object == null) { return; }
	long datetime = object.longValue();

	// Construct a Calendar
	GregorianCalendar calendar = new GregorianCalendar();
	calendar.setTime(new Date(datetime) );

	// Get the date/time components
	int year = calendar.get(GregorianCalendar.YEAR);
	int month = calendar.get(GregorianCalendar.MONTH) + 1;
	int day = calendar.get(GregorianCalendar.DAY_OF_MONTH);
	int hour = calendar.get(GregorianCalendar.HOUR_OF_DAY);
	int minutes = calendar.get(GregorianCalendar.MINUTE);
	int seconds = calendar.get(GregorianCalendar.SECOND);
	int millis = calendar.get(GregorianCalendar.MILLISECOND);

	// Construct the DICOM date string
	String date = "" + year;
	while ( date.length() < 4 ) { date = "0" + date; }

	if (month < 10) { date = date + "0" + month; }
	else { date = date + month; }

	if (day < 10) { date = date + "0" + day; }
	else { date = date + day; }

	// Construct the DICOM time string
	String time = "";
	if (hour < 10) { time = time + "0" + hour; }
	else { time = time + hour; }

	if (minutes < 10) { time = time + "0" + minutes; }
	else { time = time + minutes; }

	if (seconds < 10) { time = time + "0" + seconds; }
	else { time = time + seconds; }

	if (millis < 10)       { time = time + ".00" + millis; }
	else if (millis < 100) { time = time + ".0" + millis; }
	else { time = time + "." + millis; }

	// Set the Output Nodes
	_setOutputValue(DATE_NAME, date);
	_setOutputValue(TIME_NAME, time);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }
}
