/*
  COPYRIGHT NOTICE
  Copyright (c) 2007  Scott C. Neu and Arthur W. Toga

  See README.license for license notices.
*/

package edu.ucla.loni.debabel.processors.scott.dicom;

import edu.ucla.loni.dataflow.MetadataTree;
import edu.ucla.loni.dataflow.MetadataTreeArray;
import edu.ucla.loni.dataflow.MetadataTreeInputNode;
import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.DuplicateElementException;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.Vector;

/**
 * Processor that combines two DICOM data trees into one.
 *
 * @version 13 December 2007
 */
public class DicomTreeCombiner extends Processor implements Comparator
{
    /** Description of this Processor. */
    private static String DESC = "Combines two DICOM data trees into one.";

    /** Input Node name for the first input tree. */
    private static String INPUT_TREE1_NAME = "DICOM Tree 1";

    /** Input Node description for the first input tree. */
    private static String INPUT_TREE1_DESC = "First DICOM data tree.";

    /** Input Node name for the second input tree. */
    private static String INPUT_TREE2_NAME = "DICOM Tree 2";

    /** Input Node description for the second input tree. */
    private static String INPUT_TREE2_DESC = "Second DICOM data tree.";

    /** Output Node name for the combined tree. */
    private static String COMBINED_TREE_NAME = "Combined DICOM Tree";

    /** Output Node description for the combined tree. */
    private static String COMBINED_TREE_DESC =
	"First and second trees combined.";

    /**
     * Constructs a DICOM Tree Combiner with default Properties.
     *
     * @param name Name of the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public DicomTreeCombiner(String name)
    {
	super(name, DESC);

	// Create the Input Node for the first and second DICOM trees
	Class type = MetadataTree.class;
	_addInputNode(new InputNode(INPUT_TREE1_NAME, INPUT_TREE1_DESC, type) );
	_addInputNode(new InputNode(INPUT_TREE2_NAME, INPUT_TREE2_DESC, type) );

	// Create the Output Node for the combined DICOM tree
	type = MetadataTree.class;
	_addOutputNode( new OutputNode(COMBINED_TREE_NAME, COMBINED_TREE_DESC,
				       type) );
    }

    /**
     * Constructs a DICOM Tree Combiner using the specified Properties.
     *
     * @param name Name of the Processor.
     * @param properties Properties for the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public DicomTreeCombiner(String name, ProcessorProperty[] properties)
    {
	this(name);
    }

    /**
     * Generates values for the Output Nodes using the current values of the
     * Input Nodes.
     *
     * @throws ProcessingException If an error occurs during processing.
     */
    public void generateOutputValues() throws ProcessingException
    {
	try {

	    // Get the first DICOM tree
	    MetadataTree dcmTree1 =
		(MetadataTree)_getInputValue(INPUT_TREE1_NAME, null);
	    if (dcmTree1 == null) { return; }

	    // Get the second DICOM tree
	    MetadataTree dcmTree2 =
		(MetadataTree)_getInputValue(INPUT_TREE2_NAME, null);
	    if (dcmTree2 == null) { return; }

	    // Combine the Data Set trees
	    MetadataTree dsTree1=
		(MetadataTree)dcmTree1.getModules().nextElement();
	    MetadataTree dsTree2=
		(MetadataTree)dcmTree2.getModules().nextElement();
	    MetadataTree dsTree12 = _combineDataSetTrees(dsTree1, dsTree2);

	    // Construct the combined tree
	    MetadataTree combinedTree = _createShallowCopy(dcmTree1);
	    combinedTree.addModule(dsTree12);

	    // Set the combined tree
	    _setOutputValue(COMBINED_TREE_NAME, combinedTree);
	}

	catch (Exception e) {
	    String msg = "Unable to generate output values for \"" + this +
		"\".";
	    throw new ProcessingException(msg, e);
	}
    }

    /**
     * Compares its two arguments for order.  Returns a negative integer,
     * zero, or a positive integer as the first argument is less than, equal
     * to, or greater than the second.
     *
     * @param o1 First Object to be compared.
     * @param o2 Second Object to be compared.
     *
     * @return Negative integer, zero, or a positive integer as the first
     *         argument is less than, equal to, or greater than the second.
     */
    public int compare(Object o1, Object o2)
    {
	// Metadata Trees
	if (o1 instanceof MetadataTree && o2 instanceof MetadataTree) {
	    MetadataTree m1 = (MetadataTree)o1;
	    MetadataTree m2 = (MetadataTree)o2;

	    // Compare names
	    return m1.getName().compareTo( m2.getName() );
	}

	return 0;
    }

    /**
     * Indicates whether some other object is equal to this Comparator.
     *
     * @param obj Reference object with which to compare.
     *
     * @return True only if the specified object is also a Comparator and it
     *         imposes the same ordering as this Comparator.
     */
    public boolean equals(Object obj)
    {
	return super.equals(obj);
    }

    /**
     * Combines the specified Data Set trees into one Data Set tree.  If
     * necessary, the children of the combined Data Set tree are sorted.  If
     * there are two identically named child trees in each Data Set tree, only
     * the first is copied to the returned tree.
     *
     * @param dataSetTree1 First Data Set tree.
     * @param dataSetTree2 Second Data Set tree.
     *
     * @return Data Set tree resulting from the combination of the two trees.
     *
     * @throws DuplicateElementException If there is an unforeseen error.
     */
    private MetadataTree _combineDataSetTrees(MetadataTree dataSetTree1,
					      MetadataTree dataSetTree2)
	throws DuplicateElementException
    {
	// Get fully sorted copies of the Data Set trees
	MetadataTree dsTree1 = _createFullSortedCopy(dataSetTree1);
	MetadataTree dsTree2 = _createFullSortedCopy(dataSetTree2);

	// Combine the child trees
	Enumeration enum1 = dsTree1.getModules();
	Enumeration enum2 = dsTree2.getModules();
	if ( !enum1.hasMoreElements() ) { return dsTree2; }
	if ( !enum2.hasMoreElements() ) { return dsTree1; }

	// Add the child trees to an empty combined tree
	MetadataTree combinedTree = _createShallowCopy(dataSetTree1);
	MetadataTree tree1 = (MetadataTree)enum1.nextElement();
	MetadataTree tree2 = (MetadataTree)enum2.nextElement();

	while (tree1 != null && tree2 != null) {
	    int compare = compare(tree1, tree2);

	    // Tree 1 belongs before tree 2
	    if (compare < 0) {
		combinedTree.addModule(tree1);
		tree1 = null;
	    }

	    // Tree 1 and tree 2 are equal
	    else if (compare == 0) {
		InputNode vrNode = tree1.getPrimaryInputNode("vr");
		Object vr = vrNode.getValue();

		// Combine nested data sets
		if ( "SQ".equals(vr) ) {
		    combinedTree.addModule( _combineSqTrees(tree1, tree2) );
		}

		// Otherwise just use tree 1 and skip tree 2
		else { combinedTree.addModule(tree1); }

		tree1 = null;
		tree2 = null;
	    }

	    // Tree 1 belongs after tree 2
	    else {
		combinedTree.addModule(tree2);
		tree2 = null;
	    }

	    // Advance to the next child trees
	    if ( tree1 == null && enum1.hasMoreElements() ) {
		tree1 = (MetadataTree)enum1.nextElement();
	    }
	    if ( tree2 == null && enum2.hasMoreElements() ) {
		tree2 = (MetadataTree)enum2.nextElement();
	    }
	}

	// Add any uncombined trees
	if (tree1 != null) { combinedTree.addModule(tree1); }
	if (tree2 != null) { combinedTree.addModule(tree2); }

	// Add remaining child trees (one Data Set tree will be completed)
	while ( enum1.hasMoreElements() ) {
	    combinedTree.addModule( (MetadataTree)enum1.nextElement() );
	}
	while ( enum2.hasMoreElements() ) {
	    combinedTree.addModule( (MetadataTree)enum2.nextElement() );
	}

	return combinedTree;
    }

    /**
     * Combines the nested Data Sets of the two trees into one tree.
     *
     * @param tree1 First tree containing Data Sets.
     * @param tree2 Second tree containing Data Sets.
     *
     * @return Tree resulting from the combination of the nested Data Sets.
     *
     * @throws DuplicateElementException If there is an unforeseen error.
     */
    private MetadataTree _combineSqTrees(MetadataTree tree1, MetadataTree tree2)
	throws DuplicateElementException
    {
	MetadataTreeInputNode vrNode =
	    (MetadataTreeInputNode)tree1.getPrimaryInputNode("vr");

	// Create the combined tree
	MetadataTree combinedTree = _createShallowCopy(tree1);
	combinedTree.addInputNode( _createCopy(vrNode, "SQ") );

	// Get the first Data Set tree
	Enumeration enum1 = tree1.getModules();
	Object dataSetTree1 = null;
	if ( enum1.hasMoreElements() ) { dataSetTree1 = enum1.nextElement(); }

	// Get the second Data Set tree
	Enumeration enum2 = tree2.getModules();
	Object dataSetTree2 = null;
	if ( enum2.hasMoreElements() ) { dataSetTree2 = enum2.nextElement(); }

	// Each tree has one nested Data Set
	if (dataSetTree1 instanceof MetadataTree &&
	    dataSetTree2 instanceof MetadataTree)
	    {
		// Combine the Data Sets
		MetadataTree newDataSetTree =
		    _combineDataSetTrees( (MetadataTree)dataSetTree1,
					  (MetadataTree)dataSetTree2 );

		// Return the new tree
		combinedTree.addModule(newDataSetTree);
		return combinedTree;
	    }

	// Mixed Data Set tree types; return the first tree
	if ( !(dataSetTree1 instanceof MetadataTreeArray) &&
	     !(dataSetTree2 instanceof MetadataTreeArray) ) { return tree1; }

	// Each tree has an array of nested Data Sets
	MetadataTreeArray dataSetTreeArray1 = (MetadataTreeArray)dataSetTree1;
	MetadataTree[] dsTrees1 = dataSetTreeArray1.getArrayValue(0);
	MetadataTreeArray dataSetTreeArray2 = (MetadataTreeArray)dataSetTree2;
	MetadataTree[] dsTrees2 = dataSetTreeArray2.getArrayValue(0);

	// Find the smaller and larger of the tree arrays
	MetadataTree[] dsMoreTrees = dsTrees1;
	MetadataTree[] dsLessTrees = dsTrees2;
	if (dsLessTrees.length > dsMoreTrees.length) {
	    dsMoreTrees = dsTrees2;
	    dsLessTrees = dsTrees1;
	}

	// Combine Data Sets
	for (int i = 0; i < dsLessTrees.length; i++) {
	    dsMoreTrees[i] = _combineDataSetTrees(dsLessTrees[i],
						  dsMoreTrees[i]);
	}

	// Create a Data Set tree array for the combined Data Sets
	if (dsMoreTrees.length > 0) {
	    MetadataTreeArray newArray = new MetadataTreeArray(dsMoreTrees[0]);
	    newArray.setArrayValue(dsMoreTrees);
	    combinedTree.addModule(newArray);
	}

	return combinedTree;
    }

    /**
     * Creates a copy of the specified Metadata Tree Input Node and gives the
     * copy the given value.
     *
     * @param inputNode Metadata Tree Input Node to copy.
     * @param value Value to set the copy.
     *
     * @return Copy of the Metadata Tree Input Node set to the value.
     */
    private MetadataTreeInputNode _createCopy(MetadataTreeInputNode inputNode,
					      Object value)
    {
	MetadataTreeInputNode cloneNode =
	    (MetadataTreeInputNode)inputNode.clone();

	// Set the value of the clone
	cloneNode.setValue(value);

	// Return the clone
	return cloneNode;
    }

    /**
     * Creates a shallow copy (name and description only) of the specified
     * Metadata Tree.
     *
     * @param tree Metadata Tree to copy.
     *
     * @return Shallow copy of the Metadata Tree.
     */
    private MetadataTree _createShallowCopy(MetadataTree tree)
    {
	String name = tree.getName();
	String desc = tree.getDescription();
	return new MetadataTree(name, desc);
    }

    /**
     * Creates a full copy (structure plus attribute values) of the specified
     * Metadata Tree.  The child trees are sorted.
     *
     * @param tree Metadata Tree to copy.
     *
     * @return Full copy of the Metadata Tree.
     *
     * @throws DuplicateElementException If there is an unforeseen error.
     */
    private MetadataTree _createFullSortedCopy(MetadataTree tree)
	throws DuplicateElementException
    {
	MetadataTree cloneTree = (MetadataTree)tree.clone();

	// Copy the attribute values
	cloneTree.setTreeInputNodeValues(tree);

	// Make sure the children are sorted
	Enumeration enum = cloneTree.getModules();
	if ( !enum.hasMoreElements() ) { return cloneTree; }

	MetadataTree lastTree = (MetadataTree)enum.nextElement();
	while ( enum.hasMoreElements() ) {
	    MetadataTree thisTree = (MetadataTree)enum.nextElement();

	    // Out of order
	    if ( thisTree.getName().compareTo(lastTree.getName()) < 0 ) {
		return _getSortedTree(cloneTree);
	    }

	    lastTree = thisTree;
	}

	// Return the sorted copy
	return cloneTree;
    }

    /**
     * Reorders the children of the specified Metadata Tree so they are sorted.
     *
     * @param tree Metadata Tree to sort.
     *
     * @return Sorted Metadata Tree.
     *
     * @throws DuplicateElementException If there is an unforeseen error.
     */
    private MetadataTree _getSortedTree(MetadataTree tree)
	throws DuplicateElementException
    {
	// Construct the root
	MetadataTree sortedTree = _createShallowCopy(tree);

	// Transfer all the primary Nodes
	Enumeration enum = tree.getPrimaryInputNodes();
	while ( enum.hasMoreElements() ) {
	    sortedTree.addInputNode((MetadataTreeInputNode)enum.nextElement());
	}

	// Get all the child Metadata Trees
	Vector trees = new Vector();
	enum = tree.getModules();
	while ( enum.hasMoreElements() ) { trees.add(enum.nextElement()); }

	// Sort the child Metadata Trees
	Collections.sort(trees, this);

	// Transfer the sorted child Metadata Trees
	enum = trees.elements();
	while ( enum.hasMoreElements() ) {
	    sortedTree.addModule( (MetadataTree)enum.nextElement() );
	}

	// Return the sorted tree
	return sortedTree;
    }
}
