/*
COPYRIGHT NOTICE
Copyright (c) 2004  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott.dicom;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;

/**
 * Processor that takes a string and hashes it into a DICOM UID.
 *
 * @version 18 December 2004
 */
public class DicomUidFormatter extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Creates a DICOM UID from a string.";

  /** Input Node name for the input string. */
  private static String INPUT_NAME = "Input String";

  /** Input Node description for the input string. */
  private static String INPUT_DESC = "String to convert.";

  /** Output Node name for the DICOM UID. */
  private static String UID_NAME = "DICOM UID";

  /** Output Node description for the DICOM UID. */
  private static String UID_DESC = "DICOM UID created from the string.";

  /**
   * Constructs a DICOM UID Formatter with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public DicomUidFormatter(String name)
    {
      super(name, DESC);

      // Create the Input Node
      Class type = String.class;
      _addInputNode( new InputNode(INPUT_NAME, INPUT_DESC, type) );

      // Create the Output Node
      type = String.class;
      _addOutputNode(new OutputNode(UID_NAME, UID_DESC, type) );
    }

  /**
   * Constructs a DICOM UID Formatter using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public DicomUidFormatter(String name, ProcessorProperty[] properties)
    {
      this(name);
    }

  /**
   * Gets the long description.
   *
   * @return A verbose description of what the Processor does.
   */
  public String getLongDescription()
    {
      StringBuffer b = new StringBuffer();

      b.append("This processor maps an input string to a DICOM UID.  This\n");
      b.append("is done by computing a hash code for the string using:\n");
      b.append("\n");
      b.append("  hash = s[0]*31^(n-1) + s[1]*31^(n-2) + ... + s[n-1]\n");
      b.append("\n");
      b.append("  where s[i] is the ith character of the string\n");
      b.append("  and long arithmetic is used.\n");
      b.append("\n");
      b.append("The hash code is prepended to UCLA's purchased DICOM\n");
      b.append("prefix:\n");
      b.append("\n");
      b.append("  2.16.124.113543.6006");
      b.append("\n");
      b.append("with \".99\" prepended to imply a de-identified UID.");
      b.append("\n");
      b.append("The generated UID is then:");
      b.append("\n");
      b.append("  2.16.124.113543.6006.99.hash");

      return b.toString();
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the input string
	String s = (String)_getInputValue(INPUT_NAME, null);
	if (s == null) { return; }

	// Compute the hash code (from Java's String source code)
	long hashCode = 0;
	for (int i = 0; i < s.length(); i++) {
	  hashCode = 31*hashCode + s.charAt(i);
	}

	// Construct the UID (change negative sign to zero)
	String uid = "2.16.124.113543.6006.99.";
	if (hashCode < 0) { uid = uid + "0" + Math.abs(hashCode); }
	else { uid = uid + hashCode; }

	// Set the output string
	_setOutputValue(UID_NAME, uid);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }
}
