/*
  COPYRIGHT NOTICE
  Copyright (c) 2011  Scott C. Neu and Arthur W. Toga

  See README.license for license notices.
*/

package edu.ucla.loni.debabel.processors.scott.dicom;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import java.util.StringTokenizer;
import java.util.Vector;

/**
 * Processor that converts from NRRD space to DICOM space.
 *
 * @version 9 May 2011
 */
public class NrrdToDicomSpace extends Processor
{
    /** Description of this Processor. */
    private static String DESC = "Converts from NRRD space to DICOM space.";

    /** Input Node name for the image dimensions. */
    private static String SIZES_NAME = "sizes";

    /** Input Node description for the image dimensions. */
    private static String SIZES_DESC = "Image dimensions.";

    /** Input Node name for the axis labels. */
    private static String KINDS_NAME = "kinds";

    /** Input Node description for the axis labels. */
    private static String KINDS_DESC = "Axis labels.";

    /** Input Node name for the axis directions and spacings. */
    private static String DIRS_NAME = "space directions";

    /** Input Node description for the axis directions and spacings. */
    private static String DIRS_DESC = "Direction and spacing of each axis.";

    /** Input Node name for the space orientation. */
    private static String SPACE_NAME = "space";

    /** Input Node description for the space orientation. */
    private static String SPACE_DESC = "Space orientation.";

    /** Input Node name for the origin. */
    private static String ORIGIN_NAME = "space origin";

    /** Input Node description for the origin. */
    private static String ORIGIN_DESC = "Origin of the coordinate system.";

    /** Output Node name for the X coordinate of the row direction cosine. */
    private static String X_ROW_NAME = "Row Direction Cosine (X)";

    /** Output Node description for the X coord of the row direction cosine. */
    private static String X_ROW_DESC = "X coord of the row direction cosine.";

    /** Output Node name for the Y coordinate of the row direction cosine. */
    private static String Y_ROW_NAME = "Row Direction Cosine (Y)";

    /** Output Node description for the Y coord of the row direction cosine. */
    private static String Y_ROW_DESC = "Y coord of the row direction cosine.";

    /** Output Node name for the Z coordinate of the row direction cosine. */
    private static String Z_ROW_NAME = "Row Direction Cosine (Z)";

    /** Output Node description for the Z coord of the row direction cosine. */
    private static String Z_ROW_DESC = "Z coord of the row direction cosine.";

    /** Output Node name for the X coordinate of the column direction cosine. */
    private static String X_COL_NAME = "Column Direction Cosine (X)";

    /** Output Node description for X coord of the column direction cosine. */
    private static String X_COL_DESC = "X coord of the column direction " +
	"cosine.";

    /** Output Node name for the Y coordinate of the column direction cosine. */
    private static String Y_COL_NAME = "Column Direction Cosine (Y)";

    /** Output Node description for Y coord of the column direction cosine. */
    private static String Y_COL_DESC = "Y coord of the column direction " +
	"cosine.";

    /** Output Node name for the Z coordinate of the column direction cosine. */
    private static String Z_COL_NAME = "Column Direction Cosine (Z)";

    /** Output Node description for Z coord of the column direction cosine. */
    private static String Z_COL_DESC = "Z coord of the column direction " +
	"cosine.";

    /** Output Node name for the X coordinate of the stack direction cosine. */
    private static String X_STK_NAME = "Stack Direction Cosine (X)";

    /** Output Node description for X coord of the stack direction cosine. */
    private static String X_STK_DESC = "X coord of the stack direction cosine.";

    /** Output Node name for the Y coordinate of the stack direction cosine. */
    private static String Y_STK_NAME = "Stack Direction Cosine (Y)";

    /** Output Node description for Y coord of the stack direction cosine. */
    private static String Y_STK_DESC = "Y coord of the stack direction cosine.";

    /** Output Node name for the Z coordinate of the stack direction cosine. */
    private static String Z_STK_NAME = "Stack Direction Cosine (Z)";

    /** Output Node description for Z coord of the stack direction cosine. */
    private static String Z_STK_DESC = "Z coord of the stack direction cosine.";

    /** Output Node name for the row spacing. */
    private static String ROW_STEP_NAME = "Row Spacing";

    /** Output Node description for the row spacing. */
    private static String ROW_STEP_DESC = "Step size between adjacent " +
	"voxels in the row direction.";

    /** Output Node name for the column spacing. */
    private static String COL_STEP_NAME = "Column Spacing";

    /** Output Node description for the column spacing. */
    private static String COL_STEP_DESC = "Step size between adjacent " +
	"voxels in the column direction.";

    /** Output Node name for the stack spacing. */
    private static String STK_STEP_NAME = "Stack Spacing";

    /** Output Node description for the stack spacing. */
    private static String STK_STEP_DESC = "Step size between adjacent " +
	"voxels in the stack direction.";

    /** Output Node name for the row start location. */
    private static String ROW_START_NAME = "Row Start Location";

    /** Output Node description for the row start location. */
    private static String ROW_START_DESC = "Starting location along the row " +
	"direction.";

    /** Output Node name for the column start location. */
    private static String COL_START_NAME = "Column Start Location";

    /** Output Node description for the column start location. */
    private static String COL_START_DESC = "Starting location along the " +
	"column direction.";

    /** Output Node name for the stack start location. */
    private static String STK_START_NAME = "Stack Start Location";

    /** Output Node description for the stack start location. */
    private static String STK_START_DESC = "Starting location along the " +
	"stack direction.";

    /** Output Node name for the image width. */
    private static String WIDTH_NAME = "Image width";

    /** Output Node description for the image width. */
    private static String WIDTH_DESC = "Width of the images.";

    /** Output Node name for the image height. */
    private static String HEIGHT_NAME = "Image height";

    /** Output Node description for the image height. */
    private static String HEIGHT_DESC = "Height of the images.";

    /** Output Node name for the number of images in an image volume. */
    private static String NUM_IMAGES_NAME = "Image number per volume";

    /** Output Node description for the number of images in an image volume. */
    private static String NUM_IMAGES_DESC = "Number of images in an image " +
	"volume.";

    /** Output Node name for the number of image volumes. */
    private static String NUM_VOLUMES_NAME = "Number of volumes";

    /** Output Node description for the number of image volumes. */
    private static String NUM_VOLUMES_DESC = "Number of image volumes.";

    /**
     * Constructs an NRRD To DICOM Space with default Properties.
     *
     * @param name Name of the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public NrrdToDicomSpace(String name)
    {
	super(name, DESC);

	// Create the Input Nodes
	Class type = String.class;
	_addInputNode(new InputNode(SIZES_NAME, SIZES_DESC, type) );
	_addInputNode(new InputNode(KINDS_NAME, KINDS_DESC, type) );
	_addInputNode(new InputNode(DIRS_NAME, DIRS_DESC, type) );
	_addInputNode(new InputNode(SPACE_NAME, SPACE_DESC, type) );
	_addInputNode(new InputNode(ORIGIN_NAME, ORIGIN_DESC, type) );

	// Create the Output Nodes for the row quantities
	type = Double.class;
	_addOutputNode( new OutputNode(ROW_START_NAME, ROW_START_DESC, type) );
	_addOutputNode( new OutputNode(ROW_STEP_NAME, ROW_STEP_DESC, type) );

	// Create the Output Nodes for the row direction cosine
	type = Double.class;
	_addOutputNode( new OutputNode(X_ROW_NAME, X_ROW_DESC, type) );
	_addOutputNode( new OutputNode(Y_ROW_NAME, Y_ROW_DESC, type) );
	_addOutputNode( new OutputNode(Z_ROW_NAME, Z_ROW_DESC, type) );

	// Create the Output Nodes for the column quantities
	type = Double.class;
	_addOutputNode( new OutputNode(COL_START_NAME, COL_START_DESC, type) );
	_addOutputNode( new OutputNode(COL_STEP_NAME, COL_STEP_DESC, type) );

	// Create the Output Nodes for the column direction cosine
	type = Double.class;
	_addOutputNode( new OutputNode(X_COL_NAME, X_COL_DESC, type) );
	_addOutputNode( new OutputNode(Y_COL_NAME, Y_COL_DESC, type) );
	_addOutputNode( new OutputNode(Z_COL_NAME, Z_COL_DESC, type) );

	// Create the Output Nodes for the stack quantities
	type = Double.class;
	_addOutputNode( new OutputNode(STK_START_NAME, STK_START_DESC, type) );
	_addOutputNode( new OutputNode(STK_STEP_NAME, STK_STEP_DESC, type) );

	// Create the Output Nodes for the stack direction cosine
	type = Double.class;
	_addOutputNode( new OutputNode(X_STK_NAME, X_STK_DESC, type) );
	_addOutputNode( new OutputNode(Y_STK_NAME, Y_STK_DESC, type) );
	_addOutputNode( new OutputNode(Z_STK_NAME, Z_STK_DESC, type) );

	// Create the Output Nodes for the image volume properties
	type = Integer.class;
	_addOutputNode( new OutputNode(WIDTH_NAME, WIDTH_DESC, type) );
	_addOutputNode( new OutputNode(HEIGHT_NAME, HEIGHT_DESC, type) );
	_addOutputNode( new OutputNode(NUM_IMAGES_NAME, NUM_IMAGES_DESC, type));
	_addOutputNode( new OutputNode(NUM_VOLUMES_NAME, NUM_VOLUMES_DESC,
				       type) );
    }

    /**
     * Constructs an NRRD To DICOM Space using the specified Properties.
     *
     * @param name Name of the Processor.
     * @param properties Properties for the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public NrrdToDicomSpace(String name, ProcessorProperty[] properties)
    {
	this(name);
    }

    /**
     * Generates values for the Output Nodes using the current values of the
     * Input Nodes.
     *
     * @throws ProcessingException If an error occurs during processing.
     */
    public void generateOutputValues() throws ProcessingException
    {
	try {

	    // Parse the sizes
	    String value = (String)_getInputValue(SIZES_NAME, null);
	    if (value == null) { return; }
	    int[] sizes = _getSizes(value);

	    // Too many dimensions
	    if (sizes.length > 4) {
		throw new ProcessingException("Too many dimensions (>4)");
	    }

	    // Determine domain variation
	    value = (String)_getInputValue(KINDS_NAME, null);
	    if (value == null) { return; }
	    boolean varyDomainFirst = _varyDomainFirst(value);

	    // Set the image width
	    int imageWidth = sizes[1];
	    if (sizes.length <= 3) { imageWidth = sizes[0]; }
	    else if (varyDomainFirst) { imageWidth = sizes[0]; }
	    _setOutputValue(WIDTH_NAME, new Integer(imageWidth));

	    // Set the image height
	    int imageHeight = sizes[2];
	    if (sizes.length <= 3) { imageHeight = sizes[1]; }
	    else if (varyDomainFirst) { imageHeight = sizes[1]; }
	    _setOutputValue(HEIGHT_NAME, new Integer(imageHeight));

	    // Set the number of images in each image volume
	    int numPerVolume;
	    if (sizes.length == 2) { numPerVolume = 1; }
	    else if (sizes.length == 3) { numPerVolume = sizes[2]; }
	    else if (varyDomainFirst) { numPerVolume = sizes[2]; }
	    else { numPerVolume = sizes[3]; }
	    _setOutputValue(NUM_IMAGES_NAME, new Integer(numPerVolume));

	    // Set the number of image volumes
	    int numVolumes = sizes[0];
	    if (sizes.length == 2) { numVolumes = 1; }
	    else if (sizes.length == 3) { numVolumes = 1; }
	    else if (varyDomainFirst) { numVolumes = sizes[3]; }
	    _setOutputValue(NUM_VOLUMES_NAME, new Integer(numVolumes));

	    // Set the origin
	    value = (String)_getInputValue(ORIGIN_NAME, "(0,0,0)");
	    double[] origin = _getOrigin(value);
	    _setOutputValue(ROW_START_NAME, new Double(origin[0]));
	    _setOutputValue(COL_START_NAME, new Double(origin[1]));
	    _setOutputValue(STK_START_NAME, new Double(origin[2]));

	    // Get the direction components
	    value = (String)_getInputValue(DIRS_NAME, null);
	    double[] d = _getDirections(value);

	    // Set the row spacing
	    double rowSpacing = Math.sqrt(d[0]*d[0] + d[1]*d[1] + d[2]*d[2]);
	    _setOutputValue(ROW_STEP_NAME, new Double(rowSpacing));

	    // Set the column spacing
	    double colSpacing = Math.sqrt(d[3]*d[3] + d[4]*d[4] + d[5]*d[5]);
	    _setOutputValue(COL_STEP_NAME, new Double(colSpacing));

	    // Set the stack spacing
	    double stkSpacing = Math.sqrt(d[6]*d[6] + d[7]*d[7] + d[8]*d[8]);
	    _setOutputValue(STK_STEP_NAME, new Double(stkSpacing));

	    // Get the space orientation
	    String space = (String)_getInputValue(SPACE_NAME, null);
	    if (space == null) { return; }

	    // Flip direction components accordingly
	    String ras = "right-anterior-superior";
	    String las = "left-anterior-superior";
	    String lps = "left-posterior-superior";
	    if (space.startsWith(ras) || space.startsWith("RAS")) {
		d[0] = -d[0];
		d[1] = -d[1];
		d[3] = -d[3];  // Flip x,y
		d[4] = -d[4];
		d[6] = -d[6];
		d[7] = -d[7];
	    }
	    else if (space.startsWith(las) || space.startsWith("LAS")) {
		d[1] = -d[1];
		d[4] = -d[4];  // Flip y
		d[7] = -d[7];
	    }
	    else if (!space.startsWith(lps) && !space.startsWith("LPS")) {
		String msg = "Space not supported:  " + space;
		throw new ProcessingException(msg);
	    }

	    // Set the row directional cosines
	    if (rowSpacing == 0) { rowSpacing = 1; }
	    _setOutputValue(X_ROW_NAME, new Double(d[0]/rowSpacing));
	    _setOutputValue(Y_ROW_NAME, new Double(d[1]/rowSpacing));
	    _setOutputValue(Z_ROW_NAME, new Double(d[2]/rowSpacing));

	    // Set the column directional cosines 
	    if (colSpacing == 0) { colSpacing = 1; }
	    _setOutputValue(X_COL_NAME, new Double(d[3]/colSpacing));
	    _setOutputValue(Y_COL_NAME, new Double(d[4]/colSpacing));
	    _setOutputValue(Z_COL_NAME, new Double(d[5]/colSpacing));

	    // Set the stack directional cosines 
	    if (stkSpacing == 0) { stkSpacing = 1; }
	    _setOutputValue(X_STK_NAME, new Double(d[6]/stkSpacing));
	    _setOutputValue(Y_STK_NAME, new Double(d[7]/stkSpacing));
	    _setOutputValue(Z_STK_NAME, new Double(d[8]/stkSpacing));
	}

	catch (Exception e) {
	    String msg = "Unable to generate output values for \"" + this +
		"\".";
	    throw new ProcessingException(msg, e);
	}
    }

    /**
     * Gets the directions.
     *
     * @param spaceDirections NRRD space directions.
     *
     * @return Numerical components of the directions.
     */
    private double[] _getDirections(String spaceDirections)
    {
	// Parse the space directions string
	String[] stringDirs = _parse(spaceDirections, " (,)none");

	// Check the size
	if (stringDirs.length != 9) {
	    String msg = stringDirs.length + "!=9 space directions";
	    throw new IllegalArgumentException(msg);
	}

	// Convert to doubles
	double[] doubleDirs = new double[stringDirs.length];
	for (int i = 0; i < doubleDirs.length; i++) {
	    doubleDirs[i] = Double.parseDouble(stringDirs[i]);
	}

	// Return the direction components
	return doubleDirs;
    }

    /**
     * Gets the origin.
     *
     * @param spaceOrigin NRRD space origin.
     *
     * @return Origin coordinates.
     */
    private double[] _getOrigin(String spaceOrigin)
    {
	// Parse the origin string
	String[] stringCoords = _parse(spaceOrigin, " (,)");

	// Convert to integers
	double[] doubleCoords = new double[stringCoords.length];
	for (int i = 0; i < doubleCoords.length; i++) {
	    doubleCoords[i] = Double.parseDouble(stringCoords[i]);
	}

	// Return the origin coordinates
	return doubleCoords;
    }

    /**
     * Gets the size of each dimension.
     *
     * @param sizes NRRD sizes.
     *
     * @return Size of each dimension.
     */
    private int[] _getSizes(String sizes)
    {
	// Get the string sizes
	String[] stringSizes = _parse(sizes, " ");

	// Convert to integers
	int[] intSizes = new int[stringSizes.length];
	for (int i = 0; i < intSizes.length; i++) {
	    intSizes[i] = Integer.parseInt(stringSizes[i]);
	}

	// Return the sizes
	return intSizes;
    }

    /**
     * Determines if the voxels vary first by domain.
     *
     * @param kinds NRRD kinds.
     *
     * @return True if the voxels vary first by domain; false if the voxels vary
     *         first by a non-domain dimension.
     *
     * @throws IllegalArgumentException If the voxels vary by neither option.
     */
    private boolean _varyDomainFirst(String kinds)
    {
	// Get the dimension types
	String[] types = _parse(kinds, " ");

	// First type is not a domain
	if ( !_isDomain(types[0]) ) {

	    // Other types must be domains
	    for (int i = 1; i < types.length; i++) {
		if ( !_isDomain(types[i]) ) {
		    throw new IllegalArgumentException("Unsupported: " + kinds);
		}
	    }

	    // Non-domain
	    return false;
	}

	// First types are domains
	for (int i = 0; i < types.length-1; i++) {
	    if ( !_isDomain(types[i]) ) {
		throw new IllegalArgumentException("Unsupported: " + kinds);
	    }
	}

	// Domain
	return true;
    }

    /**
     * Determines if the dimension is an independent variable.
     *
     * @param kind Dimension label.
     *
     * @return True if the dimension is an independent variable; false o/w.
     */
    private static boolean _isDomain(String kind)
    {
	if ( "domain".equals(kind) ) { return true; }
	if ( "space".equals(kind) ) { return true; }
	if ( "time".equals(kind) ) { return true; }

	return false;
    }
    /**
     * Parses the string into substrings.
     *
     * @param s String to parse.
     * @param delims Delimiters.
     *
     * @return Parsed substrings.
     */
    private static String[] _parse(String s, String delim)
    {
	// Parse the string
	StringTokenizer tokenizer = new StringTokenizer(s, delim);

	// Get the substrings
	Vector substrings = new Vector();
	while ( tokenizer.hasMoreTokens() ) {
	    substrings.add( tokenizer.nextToken() );
	}

	// Return an array
	return (String[])substrings.toArray(new String[0]);
    }
}
