/*
  COPYRIGHT NOTICE
  Copyright (c) 2011  Scott C. Neu and Arthur W. Toga

  See README.license for license notices.
*/

package edu.ucla.loni.debabel.processors.scott.dicom;

import edu.ucla.loni.Conversions;
import edu.ucla.loni.dataflow.MetadataTree;
import edu.ucla.loni.dataflow.MetadataTreeArray;
import edu.ucla.loni.dataflow.MetadataTreeInputNode;
import edu.ucla.loni.dataflow.TreeArrayOutputNode;
import edu.ucla.loni.dataflow.TreeInputNode;
import edu.ucla.loni.dataflow.TreeModule;
import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.DuplicateElementException;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import java.io.BufferedReader;
import java.io.StringReader;
import java.util.Iterator;
import java.util.LinkedHashMap;

/**
 * Processor that decodes the Siemens (0029, 1020) private DICOM tag.
 *
 * @version 24 October 2011
 */
public class SiemensMosaicFieldDecoder extends Processor
{
    /** Description of this Processor. */
    private static String DESC = "Decodes the Siemens (0029, 1020) DICOM tag.";

    /** Input Node name for the value of the encoded DICOM tag. */
    private static String TAG_NAME = "DICOM Tag Value";

    /** Input Node description for the value of the encoded DICOM tag. */
    private static String TAG_DESC = "Value of the (0029,1020) DICOM tag.";

    /** Output Node name for the decoded Metadata Tree. */
    private static String TREE_NAME = "Decoded Values";

    /** Output Node description for the decoded Metadata Tree. */
    private static String TREE_DESC = "Decoded values in the (0029,1020) tag.";

    /**
     * Constructs a Siemens Mosaic Field Decoder with default Properties.
     *
     * @param name Name of the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public SiemensMosaicFieldDecoder(String name)
    {
	super(name, DESC);

	// Create the Input Node
	Class type = String.class;
	_addInputNode( new InputNode(TAG_NAME, TAG_DESC, type) );

	// Create the Output Node
	_addOutputNode( new TreeArrayOutputNode(TREE_NAME, TREE_DESC,
						_getAscconvTree()) );
    }

    /**
     * Constructs a Siemens Mosaic Field Decoder using the specified Properties.
     *
     * @param name Name of the Processor.
     * @param properties Properties for the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public SiemensMosaicFieldDecoder(String name,
				     ProcessorProperty[] properties)
    {
	this(name);
    }

    /**
     * Generates values for the Output Nodes using the current values of the
     * Input Nodes.
     *
     * @throws ProcessingException If an error occurs during processing.
     */
    public void generateOutputValues() throws ProcessingException
    {
	try {

	    // Get the encoded ASCCONV hexadecimal string
	    String hexString = (String)_getInputValue(TAG_NAME, null);
	    if (hexString == null) { return; }

	    // Create an ASCCONV Metadata Tree and set its values
	    MetadataTree ascconvTree = _getAscconvTree();
	    LinkedHashMap map = _getAscconvInfo(hexString);

	    // Nothing decoded
	    if (map.size() == 0) { return; }

	    Iterator iter = map.keySet().iterator();
	    while ( iter.hasNext() ) {
		String name = (String)iter.next();
		String value = (String)map.get(name);
		_setValues(ascconvTree, name, value);
	    }

	    // Set the output value
	    MetadataTree[] trees = new MetadataTree[1];
	    trees[0] = ascconvTree;
	    _setOutputValue(TREE_NAME, trees);
	}

	catch (Exception e) {}
    }

    /**
     * Sets the values of the given Tree using the Siemens ASCCONV name and
     * value.
     *
     * @param tree Tree to set the values in.
     * @param name Name of the Siemens ASCCONV parameter.
     * @param value Value of the Siemens ASCCONV parameter.
     */
    private void _setValues(MetadataTree tree, String name, String value)
    {
	// Input Node value (no dot)
	int dotIndex = name.indexOf(".");
	if (dotIndex == -1) {

	    // Value is an array value
	    int bracketIndex = name.indexOf("[");
	    if (bracketIndex != -1) {

		// Get the child Tree corresponding to the array
		String arrayName = name.substring(0, bracketIndex);
		TreeModule childTree = tree.getModule(arrayName);
		if (childTree != null) {
		    MetadataTreeArray array = (MetadataTreeArray)childTree;

		    // Set the value of a new Metadata Tree for the array
		    MetadataTree clone = array.createMetadataTreeClone();
		    TreeInputNode node = clone.getPrimaryInputNode(arrayName);
		    node.setValue(value);

		    // Add the clone to the current array (pad by 1 for clone)
		    MetadataTree[] valueArray = array.getArrayValue(1);
		    valueArray[array.getArraySize()] = clone;
		    array.setArrayValue(valueArray);
		}
	    }

	    // Value is a single value
	    else {

		// Find the associated Input Node
		TreeInputNode node = tree.getPrimaryInputNode(name);
		if (node != null) { node.setValue(value); }
	    }
	}

	// Is an Input Node value of a child Tree
	else {
	    String childTreeName = name.substring(0, dotIndex);
	    String subName = name.substring(dotIndex+1, name.length());

	    // Child Tree is an array of Trees
	    int bracketIndex = childTreeName.indexOf("[");
	    if (bracketIndex != -1) {

		// Get the child Tree corresponding to the array
		String arrayName = childTreeName.substring(0, bracketIndex);
		TreeModule childTree = tree.getModule(arrayName);
		if (childTree != null) {
		    MetadataTreeArray array = (MetadataTreeArray)childTree;

		    // Already exists a Metadata Tree for the value
		    int arrayIndex = _getArrayIndex(childTreeName);
		    int arraySize = array.getArraySize();
		    if (arrayIndex < arraySize) {

			// Set the value in the existing Metadata Tree
			MetadataTree elementTree =
			    array.getMetadataTree(arrayIndex);
			_setValues(elementTree, subName, value);
		    }

		    // No Metadata Tree exists for the value
		    else {

			// Set the value of a new Metadata Tree
			MetadataTree clone = array.createMetadataTreeClone();
			_setValues(clone, subName, value);

			// Add the clone to the current array (pad to fit)
			int padAmt = arrayIndex - arraySize + 1;
			MetadataTree[] valueArray = array.getArrayValue(padAmt);
			valueArray[arrayIndex] = clone;
			array.setArrayValue(valueArray);
		    }
		}
	    }

	    // Child Tree is a single Tree
	    else {

		// Set the Input Node value of the child Tree
		MetadataTree childTree =
		    (MetadataTree)tree.getModule(childTreeName);
		if (childTree != null) {
		    _setValues(childTree, subName, value);
		}
	    }
	}
    }

    /**
     * Gets the array index of a Siemens ASCCONV parameter name.
     *
     * @param name Name of the Siemens ASCCONV parameter to get the array index
     *             of.
     *
     * @return Array index or -1 if no index could be determined.
     */
    private int _getArrayIndex(String name)
    {
	try {

	    // Get the indices of the brackets
	    int leftIndex = name.indexOf("[");
	    int rightIndex = name.indexOf("]");

	    // Check for validity
	    if (leftIndex == -1 || rightIndex == -1 || leftIndex > rightIndex) {
		return -1;
	    }

	    // Get the string between the brackets
	    String mid = name.substring(leftIndex+1, rightIndex);

	    // Return the integer value
	    return Integer.parseInt(mid);
	}

	// Cannot determine array index
	catch (Exception e) { return -1; }
    }

    /**
     * Decodes the Siemens ASCCONV hexadecimal string into an array of Siemens
     * ASCCONV strings.
     *
     * @param hexString Siemens ASCCONV string of 2-digit hexadecimal numbers.
     *
     * @return ASCCONV names mapped to ASCCONV values.
     */
    private LinkedHashMap _getAscconvInfo(String hexString)
    {
	LinkedHashMap entries = new LinkedHashMap();

	try {

	    // Convert the hexadecimal String into a readable String
	    byte[] bytes = Conversions.toByteArray(hexString);
	    String ascii = new String(bytes);

	    // Search for the ASCCONV delimiting strings
	    String beginDelimiter = "### ASCCONV BEGIN";
	    String endDelimiter = "### ASCCONV END ###";
	    int startIndex = ascii.lastIndexOf(beginDelimiter);
	    int endIndex = ascii.indexOf(endDelimiter, startIndex);

	    // No ASCCONV information; try again
	    if (startIndex == -1 || endIndex == -1) {
		startIndex = ascii.indexOf(beginDelimiter);
		endIndex = ascii.indexOf(endDelimiter, startIndex);

		// No ASCCONV information
		if (startIndex == -1 || endIndex == -1) { return entries; }
	    }

	    // Find the trailing ### of the beginning delimiting string
	    int trailIndex = ascii.indexOf("###", startIndex+1);
	    if (trailIndex == -1) {
		String msg = "Unable to parse ### ASCONV BEGIN ... ###";
		throw new RuntimeException(msg);
	    }

	    // Get the ASCCONV string
	    String ascconv = ascii.substring(trailIndex+3, endIndex);

	    // Separate the names from the values
	    BufferedReader reader=new BufferedReader(new StringReader(ascconv));
	    String line = reader.readLine();
	    while (line != null) {

		// Find the "=" delimiter
		String equalDelimiter = " = ";
		int equalIndex = line.indexOf(equalDelimiter);
		if (equalIndex != -1) {

		    // Get the name and value
		    String name = line.substring(0, equalIndex).trim();
		    String value = line.substring(equalIndex +
						  equalDelimiter.length(),
						  line.length());

		    // Ignore the __attribute__.size lines
		    if ( !name.endsWith("__attribute__.size") ) {
			entries.put(name, value);
		    }
		}

		// Get the next line
		line = reader.readLine();
	    }

	    // Return the ASCCONV names and values
	    return entries;
	}

	// Not decodable
	catch (Exception e) { return entries; }
    }

    /**
     * Gets the Siemens ASCCONV Metadata Tree.
     *
     * @return Siemens ASCCONV Metadata Tree.
     */
    private static MetadataTree _getAscconvTree()
    {
	MetadataTree ascTree = new MetadataTree("ASCCONV", "");

	try {

	    // ASCCONV
	    _addInputNode("ulVersion", ascTree);
	    _addInputNode("tSequenceFileName", ascTree);
	    _addInputNode("tProtocolName", ascTree);
	    _addInputNode("tReferenceImage0", ascTree);
	    _addInputNode("tReferenceImage1", ascTree);
	    _addInputNode("tReferenceImage2", ascTree);
	    _addInputNode("ucScanRegionPosValid", ascTree);
	    _addInputNode("lContrasts", ascTree);
	    _addInputNode("lCombinedEchoes", ascTree);
	    _addInputNode("ucDisableChangeStoreImages", ascTree);
	    _addInputNode("ucReconstructionMode", ascTree);
	    _addInputNode("ucOneSeriesForAllMeas", ascTree);
	    _addInputNode("ucPHAPSMode", ascTree);
	    _addInputNode("ucDixon", ascTree);
	    _addInputNode("ucWaitForPrepareCompletion", ascTree);
	    _addInputNode("lAverages", ascTree);
	    _addInputNode("lRepetitions", ascTree);
	    _addInputNode("dFlipAngleDegree", ascTree);
	    _addInputNode("lScanTimeSec", ascTree);
	    _addInputNode("lTotalScanTimeSec", ascTree);
	    _addInputNode("dRefSNR", ascTree);
	    _addInputNode("dRefSNR_VOI", ascTree);
	    _addInputNode("tdefaultEVAProt", ascTree);
	    _addInputNode("tcurrentEVAProt", ascTree);
	    _addInputNode("lParadigmPeriodicity", ascTree);

	    // sProtConsistencyInfo
	    MetadataTree sProtTree=new MetadataTree("sProtConsistencyInfo", "");
	    _addInputNode("tBaselineString", sProtTree);
	    _addInputNode("flNominalB0", sProtTree);
	    _addInputNode("flGMax", sProtTree);
	    _addInputNode("flRiseTime", sProtTree);
	    _addInputNode("lMaximumMatrixMode", sProtTree);
	    _addInputNode("lMaximumNofRxReceiverChannels", sProtTree);
	    ascTree.addModule(sProtTree);

	    // sGRADSPEC
	    MetadataTree sGradTree = new MetadataTree("sGRADSPEC", "");
	    _addInputNode("bEddyCompensationValid", sGradTree);
	    _addInputNode("bB0CompensationValid", sGradTree);
	    _addInputNode("bCrossTermCompensationValid", sGradTree);
	    _addInputNode("lOffsetX", sGradTree);
	    _addInputNode("lOffsetY", sGradTree);
	    _addInputNode("lOffsetZ", sGradTree);
	    _addInputNode("bOffsetValid", sGradTree);
	    _addInputNode("lDelayX", sGradTree);
	    _addInputNode("lDelayY", sGradTree);
	    _addInputNode("lDelayZ", sGradTree);
	    _addInputNode("bDelayValid", sGradTree);
	    _addInputNode("flSensitivityX", sGradTree);
	    _addInputNode("flSensitivityY", sGradTree);
	    _addInputNode("flSensitivityZ", sGradTree);
	    _addInputNode("bSensitivityValid", sGradTree);
	    _addInputNode("flGSWDMinRiseTime", sGradTree);
	    _addInputNode("bShimCurrentValid", sGradTree);
	    _addInputNode("ucMode", sGradTree);
	    ascTree.addModule(sGradTree);

	    // sGRADSPEC - sEddyCompensationX
	    MetadataTree sEddyXTree=new MetadataTree("sEddyCompensationX", "");

	    MetadataTree amplTree = new MetadataTree("aflAmplitude", "");
	    _addInputNode("aflAmplitude", amplTree);
	    sEddyXTree.addModule( new MetadataTreeArray(amplTree) );

	    MetadataTree timeTree = new MetadataTree("aflTimeConstant", "");
	    _addInputNode("aflTimeConstant", timeTree);
	    sEddyXTree.addModule( new MetadataTreeArray(timeTree) );

	    sGradTree.addModule(sEddyXTree);

	    // sGRADSPEC - sEddyCompensationY
	    MetadataTree sEddyYTree=new MetadataTree("sEddyCompensationY", "");

	    amplTree = new MetadataTree("aflAmplitude", "");
	    _addInputNode("aflAmplitude", amplTree);
	    sEddyYTree.addModule( new MetadataTreeArray(amplTree) );

	    timeTree = new MetadataTree("aflTimeConstant", "");
	    _addInputNode("aflTimeConstant", timeTree);
	    sEddyYTree.addModule( new MetadataTreeArray(timeTree) );

	    sGradTree.addModule(sEddyYTree);

	    // sGRADSPEC - sEddyCompensationZ
	    MetadataTree sEddyZTree=new MetadataTree("sEddyCompensationZ", "");

	    amplTree = new MetadataTree("aflAmplitude", "");
	    _addInputNode("aflAmplitude", amplTree);
	    sEddyZTree.addModule( new MetadataTreeArray(amplTree) );

	    timeTree = new MetadataTree("aflTimeConstant", "");
	    _addInputNode("aflTimeConstant", timeTree);
	    sEddyZTree.addModule( new MetadataTreeArray(timeTree) );

	    sGradTree.addModule(sEddyZTree);

	    // sGRADSPEC - sB0CompensationX
	    MetadataTree sBoXTree = new MetadataTree("sB0CompensationX", "");

	    amplTree = new MetadataTree("aflAmplitude", "");
	    _addInputNode("aflAmplitude", amplTree);
	    sBoXTree.addModule( new MetadataTreeArray(amplTree) );

	    timeTree = new MetadataTree("aflTimeConstant", "");
	    _addInputNode("aflTimeConstant", timeTree);
	    sBoXTree.addModule( new MetadataTreeArray(timeTree) );

	    sGradTree.addModule(sBoXTree);

	    // sGRADSPEC - sB0CompensationY
	    MetadataTree sBoYTree = new MetadataTree("sB0CompensationY", "");

	    amplTree = new MetadataTree("aflAmplitude", "");
	    _addInputNode("aflAmplitude", amplTree);
	    sBoYTree.addModule( new MetadataTreeArray(amplTree) );

	    timeTree = new MetadataTree("aflTimeConstant", "");
	    _addInputNode("aflTimeConstant", timeTree);
	    sBoYTree.addModule( new MetadataTreeArray(timeTree) );

	    sGradTree.addModule(sBoYTree);

	    // sGRADSPEC - sB0CompensationZ
	    MetadataTree sBoZTree = new MetadataTree("sB0CompensationZ", "");

	    amplTree = new MetadataTree("aflAmplitude", "");
	    _addInputNode("aflAmplitude", amplTree);
	    sBoZTree.addModule( new MetadataTreeArray(amplTree) );

	    timeTree = new MetadataTree("aflTimeConstant", "");
	    _addInputNode("aflTimeConstant", timeTree);
	    sBoZTree.addModule( new MetadataTreeArray(timeTree) );

	    sGradTree.addModule(sBoZTree);

	    // sGRADSPEC - sCrossTermCompensationXY
	    MetadataTree sXYTree =
		new MetadataTree("sCrossTermCompensationXY","");

	    amplTree = new MetadataTree("aflAmplitude", "");
	    _addInputNode("aflAmplitude", amplTree);
	    sXYTree.addModule( new MetadataTreeArray(amplTree) );

	    timeTree = new MetadataTree("aflTimeConstant", "");
	    _addInputNode("aflTimeConstant", timeTree);
	    sXYTree.addModule( new MetadataTreeArray(timeTree) );

	    sGradTree.addModule(sXYTree);

	    // sGRADSPEC - sCrossTermCompensationXZ
	    MetadataTree sXZTree =
		new MetadataTree("sCrossTermCompensationXZ","");

	    amplTree = new MetadataTree("aflAmplitude", "");
	    _addInputNode("aflAmplitude", amplTree);
	    sXZTree.addModule( new MetadataTreeArray(amplTree) );

	    timeTree = new MetadataTree("aflTimeConstant", "");
	    _addInputNode("aflTimeConstant", timeTree);
	    sXZTree.addModule( new MetadataTreeArray(timeTree) );

	    sGradTree.addModule(sXZTree);

	    // sGRADSPEC - sCrossTermCompensationYX
	    MetadataTree sYXTree =
		new MetadataTree("sCrossTermCompensationYX","");

	    amplTree = new MetadataTree("aflAmplitude", "");
	    _addInputNode("aflAmplitude", amplTree);
	    sYXTree.addModule( new MetadataTreeArray(amplTree) );

	    timeTree = new MetadataTree("aflTimeConstant", "");
	    _addInputNode("aflTimeConstant", timeTree);
	    sYXTree.addModule( new MetadataTreeArray(timeTree) );

	    sGradTree.addModule(sYXTree);

	    // sGRADSPEC - sCrossTermCompensationYZ
	    MetadataTree sYZTree =
		new MetadataTree("sCrossTermCompensationYZ","");

	    amplTree = new MetadataTree("aflAmplitude", "");
	    _addInputNode("aflAmplitude", amplTree);
	    sYZTree.addModule( new MetadataTreeArray(amplTree) );

	    timeTree = new MetadataTree("aflTimeConstant", "");
	    _addInputNode("aflTimeConstant", timeTree);
	    sYZTree.addModule( new MetadataTreeArray(timeTree) );

	    sGradTree.addModule(sYZTree);

	    // sGRADSPEC - sCrossTermCompensationZX
	    MetadataTree sZXTree =
		new MetadataTree("sCrossTermCompensationZX","");

	    amplTree = new MetadataTree("aflAmplitude", "");
	    _addInputNode("aflAmplitude", amplTree);
	    sZXTree.addModule( new MetadataTreeArray(amplTree) );

	    timeTree = new MetadataTree("aflTimeConstant", "");
	    _addInputNode("aflTimeConstant", timeTree);
	    sZXTree.addModule( new MetadataTreeArray(timeTree) );

	    sGradTree.addModule(sZXTree);

	    // sGRADSPEC - alShimCurrent
	    MetadataTree alShimTree = new MetadataTree("alShimCurrent","");
	    _addInputNode("alShimCurrent", alShimTree);
	    sGradTree.addModule( new MetadataTreeArray(alShimTree) );

	    // sTXSPEC
	    MetadataTree sTxTree = new MetadataTree("sTXSPEC", "");
	    _addInputNode("lNoOfTraPulses", sTxTree);
	    _addInputNode("lBTB1ParallelCapacity", sTxTree);
	    _addInputNode("lBTB1SerialCapacity", sTxTree);
	    _addInputNode("lBTB2ParallelCapacity", sTxTree);
	    _addInputNode("lBTB2SerialCapacity", sTxTree);
	    _addInputNode("bBTBValid", sTxTree);
	    _addInputNode("flKDynMagnitudeMin", sTxTree);
	    _addInputNode("flKDynMagnitudeMax", sTxTree);
	    _addInputNode("flKDynMagnitudeClipLow", sTxTree);
	    _addInputNode("flKDynMagnitudeClipHigh", sTxTree);
	    _addInputNode("flKDynPhaseMax", sTxTree);
	    _addInputNode("flKDynPhaseClip", sTxTree);
	    _addInputNode("bKDynValid", sTxTree);
	    _addInputNode("ucRFPulseType", sTxTree);
	    _addInputNode("ucExcitMode", sTxTree);
	    _addInputNode("ucSimultaneousExcitation", sTxTree);
	    ascTree.addModule(sTxTree);

	    // sTXSPEC - asNucleusInfo
	    MetadataTree asNuclTree = new MetadataTree("asNucleusInfo", "");
	    _addInputNode("tNucleus", asNuclTree);
	    _addInputNode("lFrequency", asNuclTree);
	    _addInputNode("bFrequencyValid", asNuclTree);
	    _addInputNode("flReferenceAmplitude", asNuclTree);
	    _addInputNode("bReferenceAmplitudeValid", asNuclTree);
	    _addInputNode("flAmplitudeCorrection", asNuclTree);
	    _addInputNode("bAmplitudeCorrectionValid", asNuclTree);
	    sTxTree.addModule( new MetadataTreeArray(asNuclTree) );

	    // sTXSPEC - aRFPULSE
	    MetadataTree aRfpTree = new MetadataTree("aRFPULSE", "");
	    _addInputNode("tName", aRfpTree);
	    _addInputNode("bAmplitudeValid", aRfpTree);
	    _addInputNode("flAmplitude", aRfpTree);
	    sTxTree.addModule( new MetadataTreeArray(aRfpTree) );

	    // sRXSPEC
	    MetadataTree sRxTree = new MetadataTree("sRXSPEC", "");
	    _addInputNode("lGain", sRxTree);
	    _addInputNode("bGainValid", sRxTree);
	    _addInputNode("bVariCapVoltagesValid", sRxTree);
	    ascTree.addModule(sRxTree);

	    // sRXSPEC - aFFT_SCALE
	    MetadataTree aFftTree = new MetadataTree("aFFT_SCALE", "");
	    _addInputNode("lRxChannel", aFftTree);
	    _addInputNode("flFactor", aFftTree);
	    _addInputNode("bValid", aFftTree);
	    sRxTree.addModule( new MetadataTreeArray(aFftTree) );

	    // sRXSPEC - alDwellTime
	    MetadataTree alDTree = new MetadataTree("alDwellTime", "");
	    _addInputNode("alDwellTime", alDTree);
	    sRxTree.addModule( new MetadataTreeArray(alDTree) );

	    // sAdjFreSpec
	    MetadataTree sAdjFTree = new MetadataTree("sAdjFreSpec", "");
	    _addInputNode("ulMode", sAdjFTree);
	    ascTree.addModule(sAdjFTree);

	    // sAdjShimSpec
	    MetadataTree sAdjSTree = new MetadataTree("sAdjShimSpec", "");
	    _addInputNode("ulMode", sAdjSTree);
	    ascTree.addModule(sAdjSTree);

	    // sAdjWatSupSpec
	    MetadataTree sAdjWTree = new MetadataTree("sAdjWatSupSpec", "");
	    _addInputNode("ulMode", sAdjWTree);
	    ascTree.addModule(sAdjWTree);

	    // alTR
	    MetadataTree alTrTree = new MetadataTree("alTR", "");
	    _addInputNode("alTR", alTrTree);
	    ascTree.addModule( new MetadataTreeArray(alTrTree) );

	    // alTE
	    MetadataTree alTeTree = new MetadataTree("alTE", "");
	    _addInputNode("alTE", alTeTree);
	    ascTree.addModule( new MetadataTreeArray(alTeTree) );

	    // sSliceArray
	    MetadataTree sSliceTree = new MetadataTree("sSliceArray", "");
	    _addInputNode("lSize", sSliceTree);
	    _addInputNode("lConc", sSliceTree);
	    _addInputNode("ucMode", sSliceTree);
	    _addInputNode("ucImageNumbSag", sSliceTree);
	    _addInputNode("ucImageNumbCor", sSliceTree);
	    ascTree.addModule(sSliceTree);

	    // sSliceArray - asSlice
	    MetadataTree asSliceTree = new MetadataTree("asSlice", "");
	    _addInputNode("dThickness", asSliceTree);
	    _addInputNode("dPhaseFOV", asSliceTree);
	    _addInputNode("dReadoutFOV", asSliceTree);
	    _addInputNode("dInPlaneRot", asSliceTree);

	    MetadataTree sPosTree = new MetadataTree("sPosition", "");
	    _addInputNode("dSag", sPosTree);
	    _addInputNode("dCor", sPosTree);
	    _addInputNode("dTra", sPosTree);
	    asSliceTree.addModule(sPosTree);

	    MetadataTree sNorTree = new MetadataTree("sNormal", "");
	    _addInputNode("dSag", sNorTree);
	    _addInputNode("dCor", sNorTree);
	    _addInputNode("dTra", sNorTree);
	    asSliceTree.addModule(sNorTree);

	    sSliceTree.addModule( new MetadataTreeArray(asSliceTree) );

	    // sSliceArray - anAsc
	    MetadataTree anAscTree = new MetadataTree("anAsc", "");
	    _addInputNode("anAsc", anAscTree);
	    sSliceTree.addModule( new MetadataTreeArray(anAscTree) );

	    // sSliceArray - anPos
	    MetadataTree anPosTree = new MetadataTree("anPos", "");
	    _addInputNode("anPos", anPosTree);
	    sSliceTree.addModule( new MetadataTreeArray(anPosTree) );

	    // sSliceArray - sTSat
	    MetadataTree sTSatTree = new MetadataTree("sTSat", "");
	    _addInputNode("dThickness", sTSatTree);
	    sSliceTree.addModule(sTSatTree);

	    // sGroupArray
	    MetadataTree sGroupTree = new MetadataTree("sGroupArray", "");
	    _addInputNode("lSize", sGroupTree);
	    ascTree.addModule(sGroupTree);

	    // sGroupArray - asGroup
	    MetadataTree asGroupTree = new MetadataTree("asGroup", "");
	    _addInputNode("nSize", asGroupTree);
	    _addInputNode("dDistFact", asGroupTree);
	    sGroupTree.addModule( new MetadataTreeArray(asGroupTree) );

	    // sGroupArray - anMember
	    MetadataTree anMemberTree = new MetadataTree("anMember", "");
	    _addInputNode("anMember", anMemberTree);
	    sGroupTree.addModule( new MetadataTreeArray(anMemberTree) );

	    // sGroupArray - sPSat
	    MetadataTree sPSatTree = new MetadataTree("sPSat", "");
	    _addInputNode("dThickness", sPSatTree);
	    _addInputNode("dGap", sPSatTree);
	    sGroupTree.addModule(sPSatTree);

	    // sAutoAlign
	    MetadataTree sAutoTree = new MetadataTree("sAutoAlign", "");
	    ascTree.addModule(sAutoTree);

	    // sAutoAlign - dAAMatrix
	    MetadataTree dAATree = new MetadataTree("dAAMatrix", "");
	    _addInputNode("dAAMatrix", dAATree);
	    sAutoTree.addModule( new MetadataTreeArray(dAATree) );

	    // sNavigatorPara
	    MetadataTree sNavTree = new MetadataTree("sNavigatorPara", "");
	    _addInputNode("ucRespComp", sNavTree);
	    ascTree.addModule(sNavTree);

	    // sNavigatorPara - adFree
	    MetadataTree adFreeTree = new MetadataTree("adFree", "");
	    _addInputNode("adFree", adFreeTree);
	    sNavTree.addModule( new MetadataTreeArray(adFreeTree) );

	    // sPrepPulses
	    MetadataTree sPrepTree = new MetadataTree("sPrepPulses", "");
	    _addInputNode("ucFatSat", sPrepTree);
	    _addInputNode("ucWaterSat", sPrepTree);
	    _addInputNode("ucInversion", sPrepTree);
	    _addInputNode("ucSatRecovery", sPrepTree);
	    _addInputNode("ucFatSatMode", sPrepTree);
	    ascTree.addModule(sPrepTree);

	    // sKSpace
	    MetadataTree sKSpaceTree = new MetadataTree("sKSpace", "");
	    _addInputNode("lBaseResolution", sKSpaceTree);
	    _addInputNode("lPhaseEncodingLines", sKSpaceTree);
	    _addInputNode("dPhaseResolution", sKSpaceTree);
	    _addInputNode("lPartitions", sKSpaceTree);
	    _addInputNode("lImagesPerSlab", sKSpaceTree);
	    _addInputNode("dSliceResolution", sKSpaceTree);
	    _addInputNode("ucPhasePartialFourier", sKSpaceTree);
	    _addInputNode("ucSlicePartialFourier", sKSpaceTree);
	    _addInputNode("ucAveragingMode", sKSpaceTree);
	    _addInputNode("ucMultiSliceMode", sKSpaceTree);
	    _addInputNode("ucDimension", sKSpaceTree);
	    _addInputNode("unReordering", sKSpaceTree);
	    ascTree.addModule(sKSpaceTree);

	    // sFastImaging
	    MetadataTree sFastTree = new MetadataTree("sFastImaging", "");
	    _addInputNode("lEPIFactor", sFastTree);
	    _addInputNode("lTurboFactor", sFastTree);
	    _addInputNode("lSegments", sFastTree);
	    ascTree.addModule(sFastTree);

	    // sPhysioImaging
	    MetadataTree sPhysioTree = new MetadataTree("sPhysioImaging", "");
	    _addInputNode("lSignal1", sPhysioTree);
	    _addInputNode("lMethod1", sPhysioTree);
	    _addInputNode("lSignal2", sPhysioTree);
	    _addInputNode("lMethod2", sPhysioTree);
	    _addInputNode("lPhases", sPhysioTree);
	    _addInputNode("lRetroGatedImages", sPhysioTree);
	    ascTree.addModule(sPhysioTree);

	    // sPhysioImaging - sPhysioECG
	    MetadataTree ecgTree = new MetadataTree("sPhysioECG", "");
	    _addInputNode("lTriggerPulses", ecgTree);
	    _addInputNode("lTriggerWindow", ecgTree);
	    _addInputNode("lArrhythmiaDetection", ecgTree);
	    _addInputNode("lCardiacGateOnThreshold", ecgTree);
	    _addInputNode("lCardiacGateOffThreshold", ecgTree);
	    sPhysioTree.addModule(ecgTree);

	    // sPhysioImaging - sPhysioPulse
	    MetadataTree pulseTree = new MetadataTree("sPhysioPulse", "");
	    _addInputNode("lTriggerPulses", pulseTree);
	    _addInputNode("lTriggerWindow", pulseTree);
	    _addInputNode("lCardiacGateOnThreshold", pulseTree);
	    _addInputNode("lCardiacGateOffThreshold", pulseTree);
	    sPhysioTree.addModule(pulseTree);

	    // sPhysioImaging - sPhysioExt
	    MetadataTree extTree = new MetadataTree("sPhysioExt", "");
	    _addInputNode("lTriggerPulses", extTree);
	    _addInputNode("lTriggerWindow", extTree);
	    _addInputNode("lCardiacGateOnThreshold", extTree);
	    _addInputNode("lCardiacGateOffThreshold", extTree);
	    sPhysioTree.addModule(extTree);

	    // sPhysioImaging - sPhysioResp
	    MetadataTree respTree = new MetadataTree("sPhysioResp", "");
	    _addInputNode("lRespGateThreshold", respTree);
	    _addInputNode("lRespGatePhase", respTree);
	    _addInputNode("dGatingRatio", respTree);
	    sPhysioTree.addModule(respTree);

	    // sSpecPara
	    MetadataTree specTree = new MetadataTree("sSpecPara", "");
	    _addInputNode("lPhaseCyclingType", specTree);
	    _addInputNode("lPhaseEncodingType", specTree);
	    _addInputNode("lRFExcitationBandwidth", specTree);
	    _addInputNode("ucRemoveOversampling", specTree);
	    _addInputNode("lDecouplingType", specTree);
	    _addInputNode("lNOEType", specTree);
	    _addInputNode("lExcitationType", specTree);
	    _addInputNode("lSpectralSuppression", specTree);
	    ascTree.addModule(specTree);

	    // sDiffusion
	    MetadataTree sDiffTree = new MetadataTree("sDiffusion", "");
	    _addInputNode("lDiffWeightings", sDiffTree);
	    _addInputNode("lNoiseLevel", sDiffTree);
	    _addInputNode("lDiffDirections", sDiffTree);
	    _addInputNode("ulMode", sDiffTree);
	    _addInputNode("ucDiffWeightedImage", sDiffTree);
	    ascTree.addModule(sDiffTree);

	    // sDiffusion - alBValue
	    MetadataTree alBTree = new MetadataTree("alBValue", "");
	    _addInputNode("alBValue", alBTree);
	    sDiffTree.addModule( new MetadataTreeArray(alBTree) );

	    // sAngio
	    MetadataTree sAngioTree = new MetadataTree("sAngio", "");
	    _addInputNode("ucPCFlowMode", sAngioTree);
	    _addInputNode("ucTOFInflow", sAngioTree);
	    ascTree.addModule(sAngioTree);

	    // sEllipticalFilter
	    MetadataTree sEllTree = new MetadataTree("sEllipticalFilter", "");
	    _addInputNode("ucMode", sEllTree);
	    ascTree.addModule(sEllTree);

	    // sPat
	    MetadataTree sPatTree = new MetadataTree("sPat", "");
	    _addInputNode("lAccelFactPE", sPatTree);
	    _addInputNode("lAccelFact3D", sPatTree);
	    _addInputNode("ucPATMode", sPatTree);
	    _addInputNode("ucRefScanMode", sPatTree);
	    ascTree.addModule(sPatTree);

	    // sCOIL_SELECT_MEAS
	    MetadataTree sCoilTree = new MetadataTree("sCOIL_SELECT_MEAS", "");
	    ascTree.addModule(sCoilTree);

	    // sCOIL_SELECT_MEAS - asList
	    MetadataTree asListTree = new MetadataTree("asList", "");
	    _addInputNode("lElementSelected", asListTree);
	    _addInputNode("lRxChannelConnected", asListTree);

	    MetadataTree eleTree = new MetadataTree("sCoilElementID", "");
	    _addInputNode("tCoilID", eleTree);
	    _addInputNode("lCoilCopy", eleTree);
	    _addInputNode("tElement", eleTree);
	    asListTree.addModule(eleTree);

	    sCoilTree.addModule( new MetadataTreeArray(asListTree) );

	    // sCOIL_SELECT_MEAS - sCOILPLUGS
	    MetadataTree plugTree = new MetadataTree("sCOILPLUGS", "");
	    sCoilTree.addModule(plugTree);

	    MetadataTree aulTree = new MetadataTree("aulPlugId", "");
	    _addInputNode("aulPlugId", aulTree);
	    plugTree.addModule( new MetadataTreeArray(aulTree) );

	    MetadataTree auiTree = new MetadataTree("auiNmbrOfNibbles", "");
	    _addInputNode("auiNmbrOfNibbles", auiTree);
	    plugTree.addModule( new MetadataTreeArray(auiTree) );

	    // sCoilSelectMeas
	    MetadataTree sCoilTree3 = new MetadataTree("sCoilSelectMeas", "");
	    ascTree.addModule(sCoilTree3);

	    // sCoilSelectMeas - aRxCoilSelectData
	    MetadataTree aRxCoilTree = new MetadataTree("aRxCoilSelectData","");
	    sCoilTree3.addModule( new MetadataTreeArray(aRxCoilTree) );

	    // sCoilSelectMeas - aRxCoilSelectData - asList
	    MetadataTree asListTree3 = new MetadataTree("asList", "");
	    _addInputNode("lElementSelected", asListTree3);
	    _addInputNode("lRxChannelConnected", asListTree3);

	    MetadataTree eleTree3 = new MetadataTree("sCoilElementID", "");
	    _addInputNode("tCoilID", eleTree3);
	    _addInputNode("lCoilCopy", eleTree3);
	    _addInputNode("tElement", eleTree3);
	    asListTree3.addModule(eleTree3);

	    aRxCoilTree.addModule( new MetadataTreeArray(asListTree3) );

	    // asCoilSelectMeas
	    MetadataTree sCoilTree2 = new MetadataTree("asCoilSelectMeas", "");
	    ascTree.addModule( new MetadataTreeArray(sCoilTree2) );

	    // asCoilSelectMeas - asList
	    MetadataTree asListTree2 = new MetadataTree("asList", "");
	    _addInputNode("lElementSelected", asListTree2);
	    _addInputNode("lRxChannelConnected", asListTree2);

	    MetadataTree eleTree2 = new MetadataTree("sCoilElementID", "");
	    _addInputNode("tCoilID", eleTree2);
	    _addInputNode("lCoilCopy", eleTree2);
	    _addInputNode("tElement", eleTree2);
	    asListTree2.addModule(eleTree2);

	    sCoilTree2.addModule( new MetadataTreeArray(asListTree2) );

	    // asCoilSelectMeas - sCOILPLUGS
	    MetadataTree plugTree2 = new MetadataTree("sCOILPLUGS", "");
	    sCoilTree2.addModule(plugTree2);

	    MetadataTree aulTree2 = new MetadataTree("aulPlugId", "");
	    _addInputNode("aulPlugId", aulTree2);
	    plugTree2.addModule( new MetadataTreeArray(aulTree2) );

	    MetadataTree auiTree2 = new MetadataTree("auiNmbrOfNibbles", "");
	    _addInputNode("auiNmbrOfNibbles", auiTree2);
	    plugTree2.addModule( new MetadataTreeArray(auiTree2) );

	    // sEFISPEC
	    MetadataTree sefTree = new MetadataTree("sEFISPEC", "");
	    _addInputNode("bEFIDataValid", sefTree);
	    ascTree.addModule(sefTree);

	    // ucBOLDParadigmArray
	    MetadataTree ucBoldTree=new MetadataTree("ucBOLDParadigmArray", "");
	    _addInputNode("ucBOLDParadigmArray", ucBoldTree);
	    ascTree.addModule( new MetadataTreeArray(ucBoldTree) );

	    // sWiPMemBlock
	    MetadataTree swiTree = new MetadataTree("sWiPMemBlock", "");
	    ascTree.addModule(swiTree);

	    // sWiPMemBlock - alFree
	    MetadataTree alFreeTree = new MetadataTree("alFree", "");
	    _addInputNode("alFree", alFreeTree);
	    swiTree.addModule( new MetadataTreeArray(alFreeTree) );

	    // sWiPMemBlock - adFree
	    MetadataTree adFreeTree2 = new MetadataTree("adFree", "");
	    _addInputNode("adFree", adFreeTree2);
	    swiTree.addModule( new MetadataTreeArray(adFreeTree2) );
	}

	// Will not happen
	catch (Exception e) { e.printStackTrace(); }

	// Return the ASCCONV Metadata Tree
	return ascTree;
    }

    /**
     * Adds a Metadata Tree Input Node to the specified Metadata Tree.
     *
     * @param name Name of the Input Node.
     * @param tree Tree to add the Input Node to.
     *
     * @throws DuplicateElementException If the name of the Input Node matches
     *                                   the name of an existing Input Node.
     */
    private static void _addInputNode(String name, MetadataTree tree)
	throws DuplicateElementException
    {
	Class type = MetadataTreeInputNode.STRING_VALUE_TYPE;
	String desc = "";

	tree.addInputNode( new MetadataTreeInputNode(name, desc, type) );
    }
}
