/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott.identify;

import edu.ucla.loni.dataflow.identify.GeneralMetadataTree;
import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import java.util.Enumeration;
import java.util.Vector;

/**
 * Processor that finds all the General Metadata Trees that have a specified
 * name from an array.
 *
 * @version 29 September 2003
 */
public class TreeFinder extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Finds all the named trees from an array.";

  /** Name of the Input Node for the Tree name. */
  private static String STRING_NAME = "Tree Name";

  /** Description of the Input Node for the Tree name. */
  private static String STRING_DESC = "Name of the trees to find.";

  /** Name of the Input Node for the Tree array. */
  private static String TREES_NAME = "Tree Array";

  /** Description of the Input Node for the Tree array. */
  private static String TREES_DESC = "Array of Trees to search through.";

  /** Name of the Output Node for the array of Trees with the given name. */
  private static String NAMED_TREES_NAME = "Named Tree Array";

  /** Description of the Output Node for the Trees with the given name. */
  private static String NAMED_TREES_DESC = "Array of named Trees.";

  /**
   * Constructs a Tree Finder with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public TreeFinder(String name)
    {
      super(name, DESC);

      // Create the Input Node for the Tree name
      Class type = String.class;
      InputNode inputNode = new InputNode(STRING_NAME, STRING_DESC, type);
      _addInputNode(inputNode);

      // Create the Input Node for the Tree array
      type = GeneralMetadataTree[].class;
      inputNode = new InputNode(TREES_NAME, TREES_DESC, type);
      _addInputNode(inputNode);

      // Create the Output Node for the named Trees
      type = GeneralMetadataTree[].class;
      OutputNode oNode = new OutputNode(NAMED_TREES_NAME, NAMED_TREES_DESC,
					type);
      _addOutputNode(oNode);
    }

  /**
   * Constructs a Tree Finder using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public TreeFinder(String name, ProcessorProperty[] properties)
    {
      this(name);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the Tree name and array
	String name = (String)_getInputValue(STRING_NAME, null);
	GeneralMetadataTree[] trees =
	  (GeneralMetadataTree[])_getInputValue(TREES_NAME, null);

	// Search for the named Trees
	GeneralMetadataTree[] namedTrees = new GeneralMetadataTree[0];
	if (name != null && trees != null) {
	  Vector treesFound = new Vector();

	  // Check each Tree in the array
	  for (int i = 0; i < trees.length; i++) {
	    if ( name.equals(_getTreeName(trees[i])) ) {
	      treesFound.add(trees[i]);
	    }
	  }

	  // Convert the named Trees to an array
	  namedTrees = new GeneralMetadataTree[ treesFound.size() ];
	  Enumeration enum = treesFound.elements();
	  for (int i = 0; enum.hasMoreElements(); i++) {
	    namedTrees[i] = (GeneralMetadataTree)enum.nextElement();
	  }
	}

	// Set the named Trees
	_setOutputValue(NAMED_TREES_NAME, namedTrees);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the name of the General Metadata Tree.
   *
   * @param tree General Metadata Tree to get the name of.
   *
   * @return Name of the General Metadata Tree.
   */
  private String _getTreeName(GeneralMetadataTree tree)
    {
      Enumeration enum = tree.getOutputNodes();
      while ( enum.hasMoreElements() ) {
	OutputNode outputNode = (OutputNode)enum.nextElement();

	// Assume the name is the only String value
	Object value = outputNode.getValue();
	if (value instanceof String) { return (String)value; }
      }

      // Error in logic
      return null;
    }
}
