/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott.math;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.util.Vector;

/**
 * Processor that rounds a Number to an Integer.
 *
 * @version 15 September 2003
 */
public class Round extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Rounds a number to an integer.";

  /** Input Node name for the number. */
  private static String NUMBER_NAME = "Number";

  /** Input Node descripiton for the number. */
  private static String NUMBER_DESC = "Number to round.";

  /** Output Node name for the integer. */
  private static String INTEGER_NAME = "Rounded Number";

  /** Output Node description for the integer. */
  private static String INTEGER_DESC = "Number rounded to an integer.";

  /** Property name for the type of rounding. */
  private static String ROUND_NAME = "Round Algorithm";

  /** Property description for the type of rounding. */
  private static String ROUND_DESC = "Algorithm used for rounding.";

/**
   * Constructs a Round with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public Round(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a Round using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public Round(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Create the Input Node
      Class type = Number.class;
      InputNode inputNode = new InputNode(NUMBER_NAME, NUMBER_DESC, type);
      _addInputNode(inputNode);

      // Create the Output Node
      type = Integer.class;
      OutputNode outputNode = new OutputNode(INTEGER_NAME, INTEGER_DESC, type);
      _addOutputNode(outputNode);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the number to round
	Number number = (Number)_getInputValue(NUMBER_NAME, null);
	if (number == null) { return; }

	// Get the rounding algorithm
	String alg = (String)_getPropertyValue(ROUND_NAME);

	// Round the number according to the algorithm
	double d = number.doubleValue();
	int i = 0;
	if ( alg.equals("round up") ) { i = (int)Math.ceil(d); }
	if ( alg.equals("round down") ) { i = (int)Math.floor(d); }
	if ( alg.equals("round to nearest") ) { i = (int)Math.rint(d); }

	// Set the rounded number
	_setOutputValue(INTEGER_NAME, new Integer(i));
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[1];

      // Allowed rounding algorithms
      Vector optionSet = new Vector();
      optionSet.add("round up");
      optionSet.add("round to nearest");
      optionSet.add("round down");

      // Create a set Restriction
      String name = "Allowed Rounding Algorithms";
      String desc = "Algorithms that can be used for rounding.";
      Restriction restriction = new SetRestriction(name, desc, optionSet);

      // Rounding algorithm Property
      name = ROUND_NAME;
      desc = ROUND_DESC;
      Object value = "round to nearest";
      properties[0] = new ProcessorProperty(name, desc, value, restriction);

      return properties;
    }
}
