/*
  COPYRIGHT NOTICE
  Copyright (c) 2011  Scott C. Neu and Arthur W. Toga

  See README.license for license notices.
*/

package edu.ucla.loni.debabel.processors.scott.talairach;

import edu.ucla.loni.dataflow.translate.image.BufferedImageAccessor;
import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.loop.Stateful;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import java.awt.image.BufferedImage;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Vector;

/**
 * Stateful Processor that sorts DICOM images.
 *
 * @version 1 September 2011
 */
public class DicomImageSorter3 extends Processor implements Stateful
{
    /** Description of this Processor. */
    private static String DESC = "Sorts DICOM images.";

    /** Input Node name for the scale factor. */
    private static String SCALE_NAME = "Scale Factor";

    /** Input Node description for the scale factor. */
    private static String SCALE_DESC = "Factor to multiply pixels by.";

    /** Input Node name for the extra coordinate to sort by. */
    private static String T_NAME = "Volume coordinate";

    /** Input Node description for the extra coordinate to sort by. */
    private static String T_DESC = "Coordinate to sort volumes by.";

    /** Input Node name for the X coordinate of the first image pixel. */
    private static String X_NAME = "Image Position (X)";

    /** Input Node description for the X coordinate of the first image pixel. */
    private static String X_DESC = "X coordinate of the first image pixel.";

    /** Input Node name for the Y coordinate of the first image pixel. */
    private static String Y_NAME = "Image Position (Y)";

    /** Input Node description for the Y coordinate of the first image pixel. */
    private static String Y_DESC = "Y coordinate of the first image pixel.";

    /** Input Node name for the Z coordinate of the first image pixel. */
    private static String Z_NAME = "Image Position (Z)";

    /** Input Node description for the Z coordinate of the first image pixel. */
    private static String Z_DESC = "Z coordinate of the first image pixel.";

    /** Input Node name for the X coordinate of the sorting direction cosine. */
    private static String SORT_X_NAME = "Sorting Direction Cosine (X)";

    /** Input Node desc for the X coord of the sorting direction cosine. */
    private static String SORT_X_DESC =
	"Direction cosine (along X) of the vector to sort the images along.";

    /** Input Node name for the Y coordinate of the sorting direction cosine. */
    private static String SORT_Y_NAME = "Sorting Direction Cosine (Y)";

    /** Input Node desc for the Y coord of the sorting direction cosine. */
    private static String SORT_Y_DESC =
	"Direction cosine (along Y) of the vector to sort the images along.";

    /** Input Node name for the Z coordinate of the sorting direction cosine. */
    private static String SORT_Z_NAME = "Sorting Direction Cosine (Z)";

    /** Input Node desc for the Z coord of the sorting direction cosine. */
    private static String SORT_Z_DESC =
	"Direction cosine (along Z) of the vector to sort the images along.";

    /** Input Node name for the image accessor. */
    private static String ACCESSOR_NAME = "Accessor";
  
    /** Input Node description for the image accessor. */
    private static String ACCESSOR_DESC = "Accessor to sort.";

    /** Output Node name for the sorted image accessor array. */
    private static String ARRAY_NAME = "Sorted Image Accessor Array.";

    /** Output Node description for the sorted image accessor array. */
    private static String ARRAY_DESC = "Image Accessors sorted by DICOM field.";

    /** Output Node name for the sorted scale factors. */
    private static String SCALE_ARRAY_NAME = "Sorted Scale Factor Array";

    /** Output Node description for the sorted scale factors. */
    private static String SCALE_ARRAY_DESC = "Scale factors sorted by DICOM " +
	"field.";

    /** Output Node name for the average slice spacing. */
    private static String Z_SPACING_NAME = "Average Slice Spacing";

    /** Output Node description for the average slice spacing. */
    private static String Z_SPACING_DESC = "Average spacing between slices.";

    /** Output Node name for the average volume coordinate spacing. */
    private static String T_SPACING_NAME = "Average Volume Spacing";

    /** Output Node description for the average volume coordinate spacing. */
    private static String T_SPACING_DESC = "Average spacing between volumes.";

    /** Output Node name for X coordinate of the first sorted image pixel. */
    private static String ORIGIN_X_NAME = "Origin (X)";

    /** Output Node desc for X coordinate of the first sorted image pixel. */
    private static String ORIGIN_X_DESC =
	"X coordinate of the first sorted image's pixel.";

    /** Output Node name for Y coordinate of the first sorted image pixel. */
    private static String ORIGIN_Y_NAME = "Origin (Y)";

    /** Output Node desc for Y coordinate of the first sorted image pixel. */
    private static String ORIGIN_Y_DESC =
	"Y coordinate of the first sorted image's pixel.";

    /** Output Node name for Z coordinate of the first sorted image pixel. */
    private static String ORIGIN_Z_NAME = "Origin (Z)";

    /** Output Node desc for Z coordinate of the first sorted image pixel. */
    private static String ORIGIN_Z_DESC =
	"Z coordinate of the first sorted image's pixel.";

    /** Sorted DICOM Image Volumes. */
    private Vector _sortedVolumes;

    /** Coordinate to sort along. */
    private Double[] _sortingCoordinate;

    /** Dicom Image whose top left coordinates represent the origin. */
    private DicomImage _originImage;

    /**
     * Constructs a Dicom Image Sorter 2 with default Properties.
     *
     * @param name Name of the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public DicomImageSorter3(String name)
    {
	super(name, DESC);

	// Create the Input Node for the scale factor
	Class type = Double.class;
	InputNode inputNode = new InputNode(SCALE_NAME, SCALE_DESC, type);
	_addInputNode(inputNode);

	// Create the Input Node for the volume coordinate
	type = Double.class;
	inputNode = new InputNode(T_NAME, T_DESC, type);
	_addInputNode(inputNode);

	// Create the Input Node for the X coordinate of the first image pixel
	type = Double.class;
	inputNode = new InputNode(X_NAME, X_DESC, type);
	_addInputNode(inputNode);

	// Create the Input Node for the Y coordinate of the first image pixel
	type = Double.class;
	inputNode = new InputNode(Y_NAME, Y_DESC, type);
	_addInputNode(inputNode);

	// Create the Input Node for the Z coordinate of the first image pixel
	type = Double.class;
	inputNode = new InputNode(Z_NAME, Z_DESC, type);
	_addInputNode(inputNode);

	// Create the Input Node for x of the coordinate to sort along
	type = Double.class;
	inputNode = new InputNode(SORT_X_NAME, SORT_X_DESC, type);
	_addInputNode(inputNode);

	// Create the Input Node for x of the coordinate to sort along
	type = Double.class;
	inputNode = new InputNode(SORT_Y_NAME, SORT_Y_DESC, type);
	_addInputNode(inputNode);
      
	// Create the Input Node for x of the coordinate to sort along
	type = Double.class;
	inputNode = new InputNode(SORT_Z_NAME, SORT_Z_DESC, type);
	_addInputNode(inputNode);
      
	// Create the Input Node for the image
	type = BufferedImageAccessor.class;
	inputNode = new InputNode(ACCESSOR_NAME, ACCESSOR_DESC, type);
	_addInputNode(inputNode);

	// Create the Output Node for the sorted image array
	type = BufferedImageAccessor[][].class;
	OutputNode node = new OutputNode(ARRAY_NAME, ARRAY_DESC, type);
	_addOutputNode(node);

	// Create the Output Node for the sorted scale factors
	type = Double[][].class;
	node = new OutputNode(SCALE_ARRAY_NAME, SCALE_ARRAY_DESC, type);
	_addOutputNode(node);

	// Create the Output Node for the average slice spacing
	type = Double.class;
	node = new OutputNode(Z_SPACING_NAME, Z_SPACING_DESC, type);
	_addOutputNode(node);

	// Create the Output Node for the average volume spacing
	type = Double.class;
	node = new OutputNode(T_SPACING_NAME, T_SPACING_DESC, type);
	_addOutputNode(node);

	// Create the Output Node for the X coordinate of the origin
	type = Double.class;
	node = new OutputNode(ORIGIN_X_NAME, ORIGIN_X_DESC, type);
	_addOutputNode(node);

	// Create the Output Node for the Y coordinate of the origin
	type = Double.class;
	node = new OutputNode(ORIGIN_Y_NAME, ORIGIN_Y_DESC, type);
	_addOutputNode(node);

	// Create the Output Node for the Z coordinate of the origin
	type = Double.class;
	node = new OutputNode(ORIGIN_Z_NAME, ORIGIN_Z_DESC, type);
	_addOutputNode(node);

	// Clear the sorted images
	reinitializeState();
    }

    /**
     * Constructs a Dicom Image Sorter using the specified Properties.
     *
     * @param name Name of the Processor.
     * @param properties Properties for the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public DicomImageSorter3(String name, ProcessorProperty[] properties)
    {
	this(name);
    }

    /**
     * Generates values for the Output Nodes using the current values of the
     * Input Nodes.
     *
     * @throws ProcessingException If an error occurs during processing.
     */
    public void generateOutputValues() throws ProcessingException
    {
	try {

	    // Get the Input values
	    Double x = (Double)_getInputValue(X_NAME, null);
	    Double y = (Double)_getInputValue(Y_NAME, null);
	    Double z = (Double)_getInputValue(Z_NAME, null);
	    Double t = (Double)_getInputValue(T_NAME, new Double(0));
	    Double s = (Double)_getInputValue(SCALE_NAME, new Double(0));

	    // Get the sorting coordinate
	    _sortingCoordinate = new Double[3];
	    _sortingCoordinate[0] = (Double)_getInputValue(SORT_X_NAME, null);
	    _sortingCoordinate[1] = (Double)_getInputValue(SORT_Y_NAME, null);
	    _sortingCoordinate[2] = (Double)_getInputValue(SORT_Z_NAME, null);

	    if (_sortingCoordinate[0] == null) {
		String msg = "No X coordinate to sort along.";
		throw new ProcessingException(msg);
	    }
	    if (_sortingCoordinate[1] == null) {
		String msg = "No Y coordinate to sort along.";
		throw new ProcessingException(msg);
	    }
	    if (_sortingCoordinate[2] == null) {
		String msg = "No Z coordinate to sort along.";
		throw new ProcessingException(msg);
	    }

	    // Get the accessor
	    BufferedImageAccessor accessor =
		(BufferedImageAccessor)_getInputValue(ACCESSOR_NAME, null);
	    if (accessor == null) {
		String msg = "No image to sort.";
		throw new ProcessingException(msg);
	    }

	    // Construct a Dicom Image from the Input values	
	    DicomImage dicomImage = new DicomImage(x, y, z, s, accessor);

	    // Set the origin
	    if (_originImage == null) { _originImage = dicomImage; }

	    // Get the volume for the image
	    DicomImageVolume volume = null;
	    Iterator iter = _sortedVolumes.iterator();
	    while (volume == null && iter.hasNext()) {
		DicomImageVolume v = (DicomImageVolume)iter.next();
		if ( v.getT().equals(t) ) { volume = v; }
	    }

	    // Or create a new volume for the image
	    if (volume == null) {
		volume = new DicomImageVolume(t);
		_sortedVolumes.add(volume);
		Collections.sort(_sortedVolumes, _getVolumeComparator());
	    }

	    // Add the image to the volume
	    volume.addImage(dicomImage);

	    // Create the array of sorted image accessors
	    int nVol = _sortedVolumes.size();
	    BufferedImageAccessor[][] accessors =
		new BufferedImageAccessor[nVol][];
	    for (int i = 0; i < nVol; i++) {
		DicomImageVolume v = (DicomImageVolume)_sortedVolumes.get(i);
		accessors[i] = v.getAccessors();
	    }
	    _setOutputValue(ARRAY_NAME, accessors);

	    // Create the array of sorted scale factors
	    Double[][] factors = new Double[nVol][];
	    for (int i = 0; i < nVol; i++) {
		DicomImageVolume v = (DicomImageVolume)_sortedVolumes.get(i);
		factors[i] = v.getScaleFactors();
	    }
	    _setOutputValue(SCALE_ARRAY_NAME, factors);

	    // Average the slice spacings
	    double sum = 0;
	    for (int i = 0; i < nVol; i++) {
		DicomImageVolume v = (DicomImageVolume)_sortedVolumes.get(i);
		sum += v.getAverageSliceSpacing();
	    }
	    double sliceSpacing = sum/nVol;
	    _setOutputValue(Z_SPACING_NAME, new Double(sliceSpacing));

	    // Average the volume spacings
	    sum = 0;
	    for (int i = 0; i < nVol-1; i++) {
		DicomImageVolume v1 = (DicomImageVolume)_sortedVolumes.get(i);
		DicomImageVolume v2 = (DicomImageVolume)_sortedVolumes.get(i+1);
		sum += v2.getT().doubleValue() - v1.getT().doubleValue();
	    }
	    double volumeSpacing = 0;
	    if (nVol > 1) { volumeSpacing = sum/(nVol-1); }
	    _setOutputValue(T_SPACING_NAME, new Double(volumeSpacing));

	    // Set the origin using the coordinates of the first image
	    DicomImageVolume div = (DicomImageVolume)_sortedVolumes.get(0);
	    DicomImage firstImage = div.getFirstImage();
	    _setOutputValue(ORIGIN_X_NAME, firstImage.getX());
	    _setOutputValue(ORIGIN_Y_NAME, firstImage.getY());
	    _setOutputValue(ORIGIN_Z_NAME, firstImage.getZ());
	}

	catch (Exception e) {
	    String msg="Unable to generate output values for \"" + this + "\".";
	    throw new ProcessingException(msg, e);
	}
    }

    /**
     * Reinitializes the state.  This is called by a Loop Module before a loop
     * is started.
     */
    public void reinitializeState()
    {
	_sortedVolumes = new Vector();
	_originImage = null;
    }

    /**
     * Gets a Comparator for sorting DICOM image volumes.
     *
     * @return Comparator for sorting DICOM image volumes.
     */
    private Comparator _getVolumeComparator()
    {
	return new Comparator()
	    {
		/**
		 * Compares its two arguments for order.
		 *
		 * @param o1 The first object to be compared.
		 * @param o2 The second object to be compared.
		 *
		 * @return A negative integer, zero, or a positive integer as
		 *         the first argument is less than, equal to, or
		 *         greater than the second.
		 *
		 * @throws IllegalArgumentException If there isn't enough
		 *                                  information to compare them.
		 */
		public int compare(Object o1, Object o2)
		{
		    // Order by volume coordinate
		    Double t1 = ((DicomImageVolume)o1).getT();
		    Double t2 = ((DicomImageVolume)o2).getT();
		    return t1.compareTo(t2);
		}
      
		/**
		 * Indicates whether some object is equal to this Comparator.
		 *
		 * @param obj The reference object with which to compare.
		 *
		 * @return True only if the specified object is also a
		 *         comparator and it imposes the same ordering as this
		 *         comparator.
		 */
		public boolean equals(Object obj)
		{
		    return super.equals(obj);
		}
	    };
    }

    /** Image Volume. */
    private class DicomImageVolume
    {
	/** Volume coordinate. */
	private Double _t;

	/** Images in the volume. */
	private Vector _images;

	/**
	 * Constructs a DICOM Image Volume.
	 *
	 * @param t Volume coordinate.
	 */
	public DicomImageVolume(Double t)
	{
	    _t = t;
	    _images = new Vector();
	}

	/**
	 * Gets the volume coordinate.
	 *
	 * @return Volume coordinate.
	 */
	public Double getT()
	{
	    return _t;
	}

	/**
	 * Adds an image to the volume.
	 *
	 * @param image Image to add.
	 */
	public void addImage(DicomImage image)
	{
	    // Add the image
	    _images.add(image);

	    // Sort the images
	    Collections.sort(_images, _getImageComparator());
	}

	/**
	 * Gets the first sorted image in the volume.
	 *
	 * @return First sorted image in the volume.
	 */
	public DicomImage getFirstImage()
	{
	    return (DicomImage)_images.get(0);
	}

	/**
	 * Gets the image accessors in the volume.
	 *
	 * @return Image accessors in the volume.
	 */
	public BufferedImageAccessor[] getAccessors()
	{
	    BufferedImageAccessor[] accessors =
		new BufferedImageAccessor[_images.size()];

	    for (int i = 0; i < accessors.length; i++) {
		DicomImage image = (DicomImage)_images.get(i);
		accessors[i] = image.getAccessor();
	    }

	    return accessors;
	}

	/**
	 * Gets the scale factors of the images in the volume.
	 *
	 * @return Scale factors of the images in the volume.
	 */
	public Double[] getScaleFactors()
	{
	    Double[] factors = new Double[_images.size()];

	    for (int i = 0; i < factors.length; i++) {
		DicomImage image = (DicomImage)_images.get(i);
		factors[i] = image.getS();
	    }

	    return factors;
	}

	/**
	 * Gets the average slice spacing between the ordered DICOM images.
	 *
	 * @return Average slice spacing between the ordered DICOM images.
	 */
	public double getAverageSliceSpacing()
	{
	    double sliceSpacing = 0;

	    // Get the average slice spacing
	    int numberOfImages = _images.size();
	    for (int i = 0; i < numberOfImages-1; i++) {
		DicomImage d1 = (DicomImage)_images.elementAt(i);
		DicomImage d2 = (DicomImage)_images.elementAt(i+1);

		// Determine if the Image Position coordinates exists
		if ( d1.getX() != null && d1.getY() != null &&
		     d1.getZ() != null && d2.getX() != null && 
		     d2.getY() != null && d2.getZ() != null )
		    {
			double x1 = d1.getX().doubleValue();
			double x2 = d2.getX().doubleValue();
			double y1 = d1.getY().doubleValue();
			double y2 = d2.getY().doubleValue();
			double z1 = d1.getZ().doubleValue();
			double z2 = d2.getZ().doubleValue();

			double dsq = (x2-x1)*(x2-x1) + (y2-y1)*(y2-y1) +
			    (z2-z1)*(z2-z1);
			sliceSpacing += Math.sqrt(dsq);
		    }
	    }

	    // Return the average slice spacing
	    int numberOfSpaces = numberOfImages - 1;
	    if (numberOfSpaces > 0) { sliceSpacing /= numberOfSpaces; }
	    return sliceSpacing;
	}
    }

    /**
     * Gets a Comparator for sorting DICOM images.
     *
     * @return Comparator for sorting DICOM images.
     */
    private Comparator _getImageComparator()
    {
	return new Comparator()
	    {
		/**
		 * Compares its two arguments for order.
		 *
		 * @param o1 The first object to be compared.
		 * @param o2 The second object to be compared.
		 *
		 * @return A negative integer, zero, or a positive integer as
		 *         the first argument is less than, equal to, or
		 *         greater than the second.
		 *
		 * @throws IllegalArgumentException If there isn't enough
		 *                                  information to compare them.
		 */
		public int compare(Object o1, Object o2)
		{
		    // Determine if the Image Position coordinates exist
		    try { 
			double m1 = _projectionMagnitude((DicomImage)o1);
			double m2 = _projectionMagnitude((DicomImage)o2);
	      
			// *****
			//
			// DICOM STANDARDS:
			//
			// The x-y-z coordinate system on a patient is defined
			// so the x axis points to the patient's left, the y
			// axis points to the patient's posterior, and the z
			// axis points to the patient's head:
			//
			//    (*+*)     ^ Z
			//     _|_      |
			//      |       | Y
			//     / \      x -- > X
			// *****
		
			// Order along increasing coordinates
			if (m1 < m2) { return -1; }
			if (m1 == m2) { return 0; }
			return 1;
		    }
		    catch (Exception e) {
			String msg = "Not enough information to sort DICOM " +
			    "images.";
			throw new IllegalArgumentException(msg);
		    }
		}
      
		/**
		 * Indicates whether some object is equal to this Comparator.
		 *
		 * @param obj The reference object with which to compare.
		 *
		 * @return True only if the specified object is also a
		 *         comparator and it imposes the same ordering as this
		 *         comparator.
		 */
		public boolean equals(Object obj)
		{
		    return super.equals(obj);
		}

		/** 
		 * Computes the magnitude of the projection of the given image's
		 * top left coordinate onto the sorting directional vector. 
		 *
		 * @param d DicomImage whose coordinates are to be projected.
		 *
		 * @return Magnitude of the projection of the given image's
		 * top left coordinate onto the sorting directional vector.
		 */
		private double _projectionMagnitude(DicomImage d)
		{
		    // Get the coordinates of the DicomImage d
		    double vx = d.getX().doubleValue();
		    double vy = d.getY().doubleValue();
		    double vz = d.getZ().doubleValue();
	      
		    // Get the origin (first image of the loop)
		    // Not necessarily the first image in sorted order
		    double ox = _originImage.getX().doubleValue();
		    double oy = _originImage.getY().doubleValue();
		    double oz = _originImage.getZ().doubleValue();
	      
		    // Get the displacement of image d w.r.t origin
		    double ovx = vx - ox;
		    double ovy = vy - oy;
		    double ovz = vz - oz;
	      
		    // Get the sorting unit vector
		    double ux = _sortingCoordinate[0].doubleValue();
		    double uy = _sortingCoordinate[1].doubleValue();
		    double uz = _sortingCoordinate[2].doubleValue();
	      
		    // Dot v with u to get the projection onto u
		    return ovx*ux + ovy*uy + ovz*uz;  
		}
	    };
    }

    /** Buffered Image Accessor with its metadata. */
    private class DicomImage
    {
	/** X coordinate of the first image pixel. */
	private Double _x;

	/** Y coordinate of the first image pixel. */
	private Double _y;

	/** Z coordinate of the first image pixel. */
	private Double _z;

	/** Scale factor applied to the image pixels. */
	private Double _s;

	/** Buffered Image Accessor. */
	private BufferedImageAccessor _accessor;

	/**
	 * Constructs a Dicom Image.
	 *
	 * @param x X coordinate of the first image pixel.
	 * @param y Y coordinate of the first image pixel.
	 * @param z Z coordinate of the first image pixel.
	 * @param s Scale factor applied to the image pixels.
	 * @param accessor Buffered Image Accessor.
	 */
	public DicomImage(Double x, Double y, Double z, Double s,
			  BufferedImageAccessor accessor)
	{
	    _x = x;
	    _y = y;
	    _z = z;
	    _s = s;
	    _accessor = accessor;
	}

	/**
	 * Gets the X coordinate.
	 *
	 * @return X coordinate of the first image pixel.
	 */
	public Double getX()
	{
	    return _x;
	}

	/**
	 * Gets the Y coordinate.
	 *
	 * @return Y coordinate of the first image pixel.
	 */
	public Double getY()
	{
	    return _y;
	}

	/**
	 * Gets the Z coordinate.
	 *
	 * @return Z coordinate of the first image pixel.
	 */
	public Double getZ()
	{
	    return _z;
	}

	/**
	 * Gets the scale factor.
	 *
	 * @return Scale factor applied to the image pixels.
	 */
	public Double getS()
	{
	    return _s;
	}

	/**
	 * Gets the Buffered Image Accessor.
	 *
	 * @return Buffered Image Accessor.
	 */
	public BufferedImageAccessor getAccessor()
	{
	    return _accessor;
	}
    }
}
