/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott.talairach;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;

/**
 * Processor that computes the cross product of two Talairach coordinates.
 *
 * @version 9 September 2003
 */
public class TalairachCrossProduct extends Processor
{
  /** Allowed Talairach space directions. */
  private static String DIRS = Talairach.POSITIVE_X + ", " + 
                               Talairach.NEGATIVE_X + ", " +
                               Talairach.POSITIVE_Y + ", " +
                               Talairach.NEGATIVE_Y + ", " +
                               Talairach.POSITIVE_Z + ", " +
                               Talairach.NEGATIVE_Z;

  /** Description of this Processor. */
  private static String DESC = "Computes the cross product of two Talairach " +
                               "coordinates.";

  /** Input Node Name for the first Talairach coordinate. */
  private static String COORD1_NAME = "First Talairach Coordinate";

  /** Input Node description for the first Talairach coordinate. */
  private static String COORD1_DESC = DIRS;

  /** Input Node Name for the second Talairach coordinate. */
  private static String COORD2_NAME = "Second Talairach Coordinate";

  /** Input Node description for the second Talairach coordinate. */
  private static String COORD2_DESC = DIRS;

  /** Output Node Name for the cross product of the Talairach coordinates. */
  private static String COORD3_NAME = "Talairach Cross Product";

  /** Output Node description for the cross product. */
  private static String COORD3_DESC = DIRS;

  /**
   * Constructs a Talairach Cross Product with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public TalairachCrossProduct(String name)
    {
      super(name, DESC);

      // Create the Input Node for the two Talairach coordinates
      Class type = String.class;
      InputNode inputNode = new InputNode(COORD1_NAME, COORD1_DESC, type);
      _addInputNode(inputNode);

      type = String.class;
      inputNode = new InputNode(COORD2_NAME, COORD2_DESC, type);
      _addInputNode(inputNode);

      // Create the Output Node for the Talairach cross product
      type = String.class;
      OutputNode node = new OutputNode(COORD3_NAME, COORD3_DESC, type);
      _addOutputNode(node);
    }

  /**
   * Constructs a Talairach Cross Product using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public TalairachCrossProduct(String name, ProcessorProperty[] properties)
    {
      this(name);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the values of the two Talairach coordinates
	String c1 = (String)_getInputValue(COORD1_NAME, null);
	String c2 = (String)_getInputValue(COORD2_NAME, null);

	// Compute the cross product (assume a right-handed coordinate system)
	String c3 = null;
	if ( Talairach.isValid(c1) && Talairach.isValid(c2) ) {
	  
	  if ( c1.equals(Talairach.POSITIVE_X) ) {
	    if ( c2.equals(Talairach.POSITIVE_Y) ) {c3 = Talairach.POSITIVE_Z;}
	    if ( c2.equals(Talairach.NEGATIVE_Y) ) {c3 = Talairach.NEGATIVE_Z;}
	    if ( c2.equals(Talairach.POSITIVE_Z) ) {c3 = Talairach.NEGATIVE_Y;}
	    if ( c2.equals(Talairach.NEGATIVE_Z) ) {c3 = Talairach.POSITIVE_Y;}
	  }

	  if ( c1.equals(Talairach.POSITIVE_Y) ) {
	    if ( c2.equals(Talairach.POSITIVE_X) ) {c3 = Talairach.NEGATIVE_Z;}
	    if ( c2.equals(Talairach.NEGATIVE_X) ) {c3 = Talairach.POSITIVE_Z;}
	    if ( c2.equals(Talairach.POSITIVE_Z) ) {c3 = Talairach.POSITIVE_X;}
	    if ( c2.equals(Talairach.NEGATIVE_Z) ) {c3 = Talairach.NEGATIVE_X;}
	  }

	  if ( c1.equals(Talairach.POSITIVE_Z) ) {
	    if ( c2.equals(Talairach.POSITIVE_X) ) {c3 = Talairach.POSITIVE_Y;}
	    if ( c2.equals(Talairach.NEGATIVE_X) ) {c3 = Talairach.NEGATIVE_Y;}
	    if ( c2.equals(Talairach.POSITIVE_Y) ) {c3 = Talairach.NEGATIVE_X;}
	    if ( c2.equals(Talairach.NEGATIVE_Y) ) {c3 = Talairach.POSITIVE_X;}
	  }

	  if ( c1.equals(Talairach.NEGATIVE_X) ) {
	    if ( c2.equals(Talairach.POSITIVE_Y) ) {c3 = Talairach.NEGATIVE_Z;}
	    if ( c2.equals(Talairach.NEGATIVE_Y) ) {c3 = Talairach.POSITIVE_Z;}
	    if ( c2.equals(Talairach.POSITIVE_Z) ) {c3 = Talairach.POSITIVE_Y;}
	    if ( c2.equals(Talairach.NEGATIVE_Z) ) {c3 = Talairach.NEGATIVE_Y;}
	  }
	  
	  if ( c1.equals(Talairach.NEGATIVE_Y) ) {
	    if ( c2.equals(Talairach.POSITIVE_X) ) {c3 = Talairach.POSITIVE_Z;}
	    if ( c2.equals(Talairach.NEGATIVE_X) ) {c3 = Talairach.NEGATIVE_Z;}
	    if ( c2.equals(Talairach.POSITIVE_Z) ) {c3 = Talairach.NEGATIVE_X;}
	    if ( c2.equals(Talairach.NEGATIVE_Z) ) {c3 = Talairach.POSITIVE_X;}
	  }

	  if ( c1.equals(Talairach.NEGATIVE_Z) ) {
	    if ( c2.equals(Talairach.POSITIVE_X) ) {c3 = Talairach.NEGATIVE_Y;}
	    if ( c2.equals(Talairach.NEGATIVE_X) ) {c3 = Talairach.POSITIVE_Y;}
	    if ( c2.equals(Talairach.POSITIVE_Y) ) {c3 = Talairach.POSITIVE_X;}
	    if ( c2.equals(Talairach.NEGATIVE_Y) ) {c3 = Talairach.NEGATIVE_X;}
	  }
	}

	// Set the cross product
	_setOutputValue(COORD3_NAME, c3);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }
}
