/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.stdlib;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import java.lang.reflect.Array;

/**
 * Processor that indexes an element of an array.
 *
 * @author Scott C. Neu 
 *
 * @version 3 June 2003
 */
public class ArrayIndexer extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Indexes an element of an array.";

  /** Input Node name for the array. */
  private static String ARRAY_NAME = "Array";

  /** Input Node description for the array. */
  private static String ARRAY_DESC = "Array of elements.";

  /** Input Node name for the array index. */
  private static String INDEX_NAME = "Array Index.";

  /** Input Node description for the array index. */
  private static String INDEX_DESC = "Index of the array element.";

  /** Output Node name for the indexed array element. */
  private static String ELEMENT_NAME = "Indexed Array Element";

  /** Output Node description for the indexed array element. */
  private static String ELEMENT_DESC = "Element that is indexed in the array.";

  /** Property for the array type. */
  private static String PROP_NAME = "type";

  /** Property description for the array type. */
  private static String PROP_DESC = "Type of array to index.";

  /**
   * Constructs an Array Indexer with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public ArrayIndexer(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs an Array Indexer using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public ArrayIndexer(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Get the array type
      Class arrayType = (Class)_getPropertyValue(PROP_NAME);

      // Create the Input Nodes
      Class type = Array.newInstance(arrayType, 0).getClass();
      InputNode inputNode = new InputNode(ARRAY_NAME, ARRAY_DESC, type);
      _addInputNode(inputNode);

      type = Integer.class;
      inputNode = new InputNode(INDEX_NAME, INDEX_DESC, type);
      _addInputNode(inputNode);

      // Create the Output Node
      OutputNode node = new OutputNode(ELEMENT_NAME, ELEMENT_DESC, arrayType);
      _addOutputNode(node);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the array
	Object array = null;
	Object arrayValue = _getInputValue(ARRAY_NAME, null);
	if (arrayValue != null && arrayValue.getClass().isArray()) {
	  array = arrayValue;
	}

	// Get the index of the element
	Integer defaultInt = new Integer(-1);
	Integer indexValue = (Integer)_getInputValue(INDEX_NAME, defaultInt);
	int index = indexValue.intValue();

	// Set the indexed array element
	Object element = null;
	if (array != null && index >= 0) {
	  element = Array.get(array, index);
	}
	_setOutputValue(ELEMENT_NAME, element);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[1];

      // Default array type
      String name = PROP_NAME;
      String desc = PROP_DESC;
      Object value = String.class;
      Restriction restriction = Utilities.getClassRestriction();
      properties[0] = new ProcessorProperty(name, desc, value, restriction);

      return properties;
    }
}
