/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.stdlib;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.RangeRestriction;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Vector;

/**
 * Processor that manipulates a List.
 *
 * @author Keith R. Ouellette 
 *
 * @version 3 June 2003
 */
public class ListManipulator extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Manipulates a List.";

  /** Input Node name of the List to manipulate. */
  private static String LIST_NAME = "list";

  /** Input Node description of the List to manipulate. */
  private static String LIST_DESC = "List to Manipulate."; 

  /** Input Node name of the List to manipulate. */
  private static String LIST_COPY_TARGET_NAME = "list target";

  /** Input Node description of the List to manipulate. */
  private static String LIST_COPY_TARGET_DESC = "List to copy into.";

  /** Input Node name of the first swapping index. */
  private static String INDEX1_NAME = "index 1";
  
  /** Input Node name of the second swapping index. */
  private static String INDEX2_NAME = "index 2";

  /** Input Node description of the method argument. */
  private static String INDEX_DESC = "index to swap.";
  
  /** Output Node name of the Object returned. */
  private static String OBJECT_NAME = "manipulated list";
 
  /** Property for the operation. */
  private static String OP_PROP_NAME = "list manipulation";

  /** Property description for the operation. */
  private static String OP_PROP_DESC = "list manipulation to execute.";
  
  /** List operation. */
  private static String REVERSE = "reverse";

  /** List operation. */
  private static String SORT = "sort";

  /** List operation. */
  private static String SWAP = "swap";

  /** List operation. */
  private static String COPY = "copy"; 

  /**
   * Constructs a ListManipulator with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public ListManipulator(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a ListManipulator using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public ListManipulator(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Get the String operation 
      String op = (String)_getPropertyValue(OP_PROP_NAME);
     
      // Create the "List to manipulate" Input Node
      Class type = List.class;
      InputNode inputNode = new InputNode(LIST_NAME, LIST_DESC, type);      
      _addInputNode(inputNode); 

      if(op.equals(SWAP)) {
	// Create the "Index1 to sawp" Input Node
	type = Integer.class;
	inputNode = new InputNode(INDEX1_NAME, INDEX_DESC, type);
	String n = "Positive integer";
	String d = "INteger must be non negative.";
	Restriction r = new RangeRestriction(n, d, 0, Integer.MAX_VALUE);
	inputNode.setRestriction(r);
	_addInputNode(inputNode);

	// Create the "Index1 to sawp" Input Node
	type = Integer.class;
	inputNode = new InputNode(INDEX2_NAME, INDEX_DESC, type);
        r = new RangeRestriction(n, d, 0, Integer.MAX_VALUE);
	inputNode.setRestriction(r);
	_addInputNode(inputNode);	
      }
      else if(op.equals(COPY)) {
	// Create the "List to copy into" Input Node
	type = List.class;
	inputNode = new InputNode(LIST_COPY_TARGET_NAME, 
				  LIST_COPY_TARGET_DESC, type);
	_addInputNode(inputNode);
      }
      // Create the Output Node
      type = List.class;
      OutputNode outputNode = new OutputNode(OBJECT_NAME, 
                                             "Manipulated list", type);
      _addOutputNode(outputNode);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {
	// Get the Collection to search
	List value = (List) _getInputValue(LIST_NAME, null);

	// Get the String operation 
	String op = (String)_getPropertyValue(OP_PROP_NAME);
       	
	if(op.equals(REVERSE)) {	
	  Collections.reverse(value);
	}
	else if(op.equals(SORT)) {	
	  Collections.sort(value);
	}
	else if(op.equals(SWAP)) {	
	  Integer ai = (Integer) _getInputValue(INDEX1_NAME, null);
	  Integer aj = (Integer) _getInputValue(INDEX2_NAME, null);
	  Collections.swap(value, ai.intValue(), aj.intValue());
	}
	else if(op.equals(COPY)) {	
	  List target = (List) _getInputValue(LIST_COPY_TARGET_NAME, null);
	  Collections.copy(value, target);
	  value = target;
	}		
	_setOutputValue(OBJECT_NAME, value);	
      }
      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[1];

      // Default boolean operation
      String name = OP_PROP_NAME;
      String desc = OP_PROP_DESC;
      Object value = SORT;
 
      // Create the Vector of supported List manipulations.
      Vector ops = new Vector();
      ops.add(REVERSE);
      ops.add(SORT);
      ops.add(SWAP);
      ops.add(COPY);
    
      // Create a Restriction to limit the allowed supported operations
      String REST_NAME = "Allowable list manipulations";
      String REST_DESC = "Allowable list manipulations for this processor";
      Restriction restriction = new SetRestriction(REST_NAME, REST_DESC, ops);

      properties[0] = new ProcessorProperty(name, desc, value, restriction);
      return properties;
    }
}
