/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.stdlib;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.RangeRestriction;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.util.Vector;

/**
 * Processor that replaces a String with another with respect to the user 
 * specified criteria.
 *
 * @author Keith R. Ouellette 
 *
 * @version 10 June 2003
 */
public class StringReplacer extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Replaces a String with another with respect" + 
                               "to the user specified criteria.";

  /** Input Node name of the String to replace. */
  private static String STRING_NAME = "string";

  /** Input Node description of the String to replace. */
  private static String STRING_DESC = "string to operate on"; 

  /** Input Node name of the String argument. */
  private static String ARG1_NAME = "arg1";

  /** Input Node description of the String argument. */
  private static String ARG1_DESC = "argument.";

  /** Input Node name of the String argument. */
  private static String ARG2_NAME = "arg2";

  /** Input Node description of the String argument. */
  private static String ARG2_DESC = "argument.";
  
  /** Output Node name of the Replacement String. */
  private static String REP_STRING_NAME = "result";

  /** Output Node description of the Replacement String. */
  private static String REP_STRING_DESC = "string returned.";
  
  /** Property for the operation. */
  private static String OP_PROP_NAME = "operation";

  /** Property description for the operation. */
  private static String OP_PROP_DESC = "String operation to execute.";
  
  /** Operation. */
  private static String SUBSTRING = "substring";
  
  /** Operation. */
  private static String REPLACE = "replace";

  /** Operation. */
  private static String REPLACE_ALL = "replace all";

  /** Operation. */
  private static String REPLACE_FIRST = "replace first";

  /** Operation. */
  private static String TO_LOWER = "to lower case";

  /** Operation. */
  private static String TO_UPPER = "to upper case";
  
  /** Operation. */
  private static String TRUNCATE = "truncate"; 

  /**
   * Constructs a StringReplacer with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public StringReplacer(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a StringReplacer using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public StringReplacer(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Get the String operation 
      String op = (String)_getPropertyValue(OP_PROP_NAME);

      // Create the String argument Input Node
      Class type = String.class;
      InputNode inputNode = new InputNode(STRING_NAME, STRING_DESC, type);
      _addInputNode(inputNode);
      
      if(op.equals(SUBSTRING)) {
	// Create the beginning index Input Node
	type = Integer.class;
	String n = "Positive Numbers";
	String d = "Only non negative Integers are allowed.";
	Restriction r = new RangeRestriction(n, d, 0, Integer.MAX_VALUE);
        inputNode = new InputNode(ARG1_NAME, "Begin index", type);
	inputNode.setRestriction(r);
	_addInputNode(inputNode);

	// Create the end index Input Node
	inputNode = new InputNode(ARG2_NAME,"End index (not including)", type);
        r = new RangeRestriction(n, d, 0, Integer.MAX_VALUE);
	inputNode.setRestriction(r);
	_addInputNode(inputNode);
      } 
      else if(op.equals(TRUNCATE)) {
	// Create the beginning index Input Node
	type = Integer.class;
	String n = "Max number of characters";
	String d = "Only non negative Integers are allowed.";
	Restriction r = new RangeRestriction(n, d, 0, Integer.MAX_VALUE);
        inputNode = new InputNode(ARG1_NAME, "Max number of characters", type);
	inputNode.setRestriction(r);
	_addInputNode(inputNode);
      }
      else if (op.equals(REPLACE)) {
	// Create the "char to replace" Input Node
	type = Character.class;
	inputNode = new InputNode(ARG1_NAME, "Character to replace", type);
	_addInputNode(inputNode);

	// Create the "replacement char" Input Node
	inputNode = new InputNode(ARG2_NAME, "Replacement Character", type);
	_addInputNode(inputNode);
      }
      else if (op.equals(REPLACE_ALL) || op.equals(REPLACE_FIRST)) {
	// Create the "Expression to replace" Input Node
	type = String.class;
	inputNode = new InputNode(ARG1_NAME, "Expression to replace", type);
	_addInputNode(inputNode);

	// Create the "Replacement expression" Input Node
	inputNode = new InputNode(ARG2_NAME, "Replacement Expression", type);
	_addInputNode(inputNode);	
      }
      else if (op.equals(TO_LOWER)) {	
      }
      else if (op.equals(TO_UPPER)) {	
      }
      // Create the Output Node
      type = String.class;
      OutputNode outputNode = new OutputNode(REP_STRING_NAME, 
					     REP_STRING_DESC, type);
      _addOutputNode(outputNode);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {
	// Get the String operation 
	String op = (String)_getPropertyValue(OP_PROP_NAME);

	String toReplace = (String)_getInputValue(STRING_NAME, "");

	String replacement = "";

	if(op.equals(SUBSTRING)) {
	  Integer start = (Integer)_getInputValue(ARG1_NAME, new Integer(0));
	  Integer end = (Integer) _getInputValue(ARG2_NAME,
					      new Integer(toReplace.length()));
	  int sIndex = start.intValue();
	  int eIndex = end.intValue();

	  replacement = toReplace.substring(sIndex, eIndex);	  	 
	}
	else if(op.equals(TRUNCATE)) {

	  Integer numChars = (Integer)_getInputValue(ARG1_NAME,new Integer(0));
	 
	  int maxNum = numChars.intValue();
	  
	  if(maxNum > toReplace.length()) { maxNum = toReplace.length(); }

	  replacement = toReplace.substring(0, maxNum);	  	 
	}
	else if (op.equals(REPLACE)) {
	  Character oldC = (Character)_getInputValue(ARG1_NAME, 
						      new Character('\0'));
	  Character newC = (Character)_getInputValue(ARG2_NAME, 
					      new Character(oldC.charValue()));
	  char oldChar = oldC.charValue();
	  char newChar = newC.charValue(); 
	  replacement = toReplace.replace(oldChar, newChar);
	}
	else if (op.equals(REPLACE_ALL)) {
	  String oldS = (String)_getInputValue(ARG1_NAME, "");
	  String newS = (String)_getInputValue(ARG2_NAME, "");
	  replacement = toReplace.replaceAll(oldS, newS);
	}
	else if (op.equals(REPLACE_FIRST)) {
	  String oldS = (String)_getInputValue(ARG1_NAME, "");
	  String newS = (String)_getInputValue(ARG2_NAME, "");
	  replacement = toReplace.replaceFirst(oldS, newS);
	}
	else if (op.equals(TO_LOWER)) {
	  replacement = toReplace.toLowerCase();
	}
	else if (op.equals(TO_UPPER)) {
	  replacement = toReplace.toUpperCase();
	}   
	_setOutputValue(REP_STRING_NAME, replacement);	
      }
      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[1];

      // Default boolean operation
      String name = OP_PROP_NAME;
      String desc = OP_PROP_DESC;
      Object value = REPLACE_ALL;

      // Create the Vector of supported String operations.
      Vector ops = new Vector();
      ops.add(SUBSTRING);
      ops.add(REPLACE);
      ops.add(REPLACE_ALL);
      ops.add(REPLACE_FIRST);
      ops.add(TO_LOWER);
      ops.add(TO_UPPER);
      ops.add(TRUNCATE);

      // Create a Restriction to limit the allowed supported operations
      String REST_NAME = "Allowable String replacements";
      String REST_DESC = "Allowable  String replacements for this processor";
      Restriction restriction = new SetRestriction(REST_NAME, REST_DESC, ops);

      properties[0] = new ProcessorProperty(name, desc, value, restriction);
      return properties;
    }
}
