/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.stdlib;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.util.Vector;

/**
 * Processor that performs the Java Ternary Operation ?: .
 *
 * @author Keith R. Ouellette
 * 
 * @version 3 June 2003
 */
public class TernaryOperator extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Performs the Java Ternary Operation ?:.";

  /** Input Node name for the Boolean. */
  private static String BOOLEAN_NAME = "boolean";

  /** Input Node description for the ith Boolean. */
  private static String BOOLEAN_DESC = "Boolean argument.";

  /** Input Node name for the Argument to return if boolean is true. */
  private static String IF_TRUE_NAME = "return if true";

  /** Input Node description for the Argument to return if boolean is true. */
  private static String IF_TRUE_DESC = "Argument to return if true";

  /** Input Node name for the Argument to return if boolean is false. */
  private static String IF_FALSE_NAME = "return if false";

  /** Input Node description for the Argument to return if boolean is false. */
  private static String IF_FALSE_DESC = "Argument to return if false";

  /** Output Node name for the resulting Boolean. */
  private static String RETURN_VALUE_NAME = "Argument returned.";

  /** Output Node description for the resulting Boolean. */
  private static String RETURN_VALUE_DESC 
                                  = "The result of the ternary operation.";
  /**
   * Constructs a TernaryOperator with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public TernaryOperator(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a TernaryOperator using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public TernaryOperator(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Create the Input Node for the boolean argument
      InputNode inputNode 
	     = new InputNode(BOOLEAN_NAME, BOOLEAN_DESC, Boolean.class);
      _addInputNode(inputNode);
       
      // Create the InputNode for the "if true return this" argument
      inputNode = new InputNode(IF_TRUE_NAME, IF_TRUE_DESC, Object.class);
      _addInputNode(inputNode);

      // Create the InputNode for the "if false return this" argument
      inputNode = new InputNode(IF_FALSE_NAME, IF_FALSE_DESC, Object.class);
      _addInputNode(inputNode);

      // Create the Output Node for the return value
      OutputNode outputNode = new OutputNode(RETURN_VALUE_NAME, 
					     RETURN_VALUE_DESC,
					     Object.class);
      _addOutputNode(outputNode);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the Boolean condition to determine what to return
	Boolean b = (Boolean)_getInputValue(BOOLEAN_NAME, new Boolean(false));
	
	boolean bool = b.booleanValue();
	
	Object returnVal = null;

	if(bool == true) {
	  returnVal = _getInputValue(IF_TRUE_NAME, null);	  
	}
	else {
	  returnVal = _getInputValue(IF_FALSE_NAME, null);
	}
	
	// Set the return value
	_setOutputValue(RETURN_VALUE_NAME, returnVal);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[0];
      return properties;
    }
}
