/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.stdlib;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import java.util.Vector;

/**
 * Processor that returns the Object's toString() value.
 *
 * @author Keith R. Ouellette 
 *
 * @version 3 June 2003
 */
public class ToString extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Returns an Object's toString() value.";

  /** Input Node Name for the Object to cast. */
  private static String OBJECT_NAME = "object";

  /** Input Node Description for the Object to cast. */
  private static String OBJECT_DESC = "Object to cast to a String.";

  /** Output Node Name for the toString() value. */
  private static String STRING_NAME = "string";

  /** Output Node Description for the toString() value. */
  private static String STRING_DESC = "Object's toString() value";

  /**
   * Constructs an ToString with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public ToString(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs an ToString using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   * @throws IllegalArgumentException If the class to cast does not exist. 
   */
  public ToString(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);
     
      // Create the Input Node
      Class type = Object.class;
      InputNode inputNode = new InputNode(OBJECT_NAME, OBJECT_DESC, type);
      _addInputNode(inputNode);

      // Create the Output Node
      type = String.class;
      OutputNode node = new OutputNode(STRING_NAME, STRING_DESC, type);
      _addOutputNode(node);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {
	// Get the obejct to cast
	Object value = _getInputValue(OBJECT_NAME, null);

	// Set the conversion
	_setOutputValue(STRING_NAME, value.toString());
      }
      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[0];
      return properties;
    }
}
