/*
  COPYRIGHT NOTICE
  Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

  See README.license for license notices.
*/

package edu.ucla.loni.debabel.processors.vann;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.loop.Stateful;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.RangeRestriction;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.awt.image.BufferedImage;
import java.lang.reflect.Array;
import java.util.Iterator;

/**
 * Processor that takes in a Long value representing the number of milli-
 * seconds since January 1, 1970, and outputs the date and time elements.
 *
 * @author Bryan C. Vann
 * 
 * @version 20 June 2003
 */
public class BufferedImageProperties extends Processor 
{  
    /** Description of this Processor. */
    private static String DESC = "Sorts a set of buffered images.";
  
    /** Base Input Node name for the image. */
    private static String IMAGE_NAME = "image";

    /** Input Node description of the current image to be sorted. */
    private static String IMAGE_DESC = "Current image to be sorted.";
				
    /** Base Output Node name of the image type. */
    private static String IMAGE_TYPE_NAME = "image type"; 

    /** Output Node description of the image type. */
    private static String IMAGE_TYPE_DESC = "Type of buffered image";
				
    /** Base Output Node name of the image type. */
    private static String BITS_PER_PIXEL_NAME = "bits per pixel"; 

    /** Output Node description of the image type. */
    private static String BITS_PER_PIXEL_DESC = "Number of bits per pixel"; 

    /**
    * Constructs a BufferedImageProperties.
    *
    * @param name Name of the Processor.
    *
    * @throws IllegalStringException If the name contains illegal characters.
    */
    public BufferedImageProperties(String name)
    {
	this(name, _getDefaultProperties());
    }

    /**
     * Constructs a BufferedImageProperties using the specified Properties.
     *
     * @param name Name of the Processor.
     * @param properties Properties for the Processor.
     *
     * @throws IllegalStringException If the name contains illegal characters.
     */
    public BufferedImageProperties(String name, ProcessorProperty[] properties)
    {
	super(name, DESC);

	// Add the Properties
	_addProperties(_getDefaultProperties(), properties);

	// Create InputNode
	name = IMAGE_NAME;
	InputNode inputNode = new InputNode(name, IMAGE_DESC, 
					    BufferedImage.class);
	_addInputNode(inputNode);

	// Create the Output Nodes	
	name = IMAGE_TYPE_NAME;
	Class type = Integer.class;
	OutputNode outputNode = new OutputNode(name, IMAGE_TYPE_DESC, type);
	_addOutputNode(outputNode);
	
	name = BITS_PER_PIXEL_NAME;
	type = Integer.class;
	outputNode = new OutputNode(name, 
				    BITS_PER_PIXEL_DESC, type);
	_addOutputNode(outputNode);
    }

    /**
     * Generates values for the Output Nodes using the current values of the
     * Input Nodes.
     *
     * @throws ProcessingException If an error occurs during processing.
     */
    public void generateOutputValues() throws ProcessingException
    {
	try {
	    // Get the index
	    BufferedImage image = 
		(BufferedImage)(_getInputValue(IMAGE_NAME, null));

	    int type = image.getType();

	    int bitsPerPixel = image.getColorModel().getPixelSize();
	    
	    // set output values
	    _setOutputValue(IMAGE_TYPE_NAME, new Integer(type));
	    _setOutputValue(BITS_PER_PIXEL_NAME, new Integer(bitsPerPixel));
	    
	}
	catch (Exception e) {
	    e.printStackTrace();
	    String msg = "Unable to generate output values for \""+this+ "\".";
	    throw new ProcessingException(msg, e);
	}
    }

    /**
     * Gets the default Properties.
     *
     * @return Default Processor Properties.
     */
    private static ProcessorProperty[] _getDefaultProperties()
    {
	ProcessorProperty[] properties = new ProcessorProperty[0];
	return properties;
    }
}
	
      

  
