/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.vann;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.RangeRestriction;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.util.Vector; 

/**
 * Processor that takes integer values of hour, minute, second, and 
 * milli-second, and formats it correctly for dicom time format.
 *
 * @author Bryan C. Vann
 * 
 * @version 24 June 2003
 */
public class DicomTimeFormatter extends Processor
{   
  /** Description of this Processor. */
  private static String DESC = "Converts time elements to dicom format.";

  /** Base Input Node name for the hour. */
  private static String HOUR_NAME = "Hour"; 

  /** Input Node description of the hour. */
  private static String HOUR_DESC = "Hour part of the time";
				
  /** Base Input Node name for the minute. */
  private static String MINUTE_NAME = "Minute"; 

  /** Input Node description of the minute. */
  private static String MINUTE_DESC = "Minute part of the time";
  
  /** Base Input Node name for the second. */
  private static String SECOND_NAME = "Second"; 

  /** Input Node description of the second. */
  private static String SECOND_DESC = "Second part of the time";
  
  /** Base Input Node name for the second. */
  private static String MILLISECOND_NAME = "Milli-second"; 

  /** Input Node description of the second. */
  private static String MILLISECOND_DESC = "Milli-second part of the time";

  /** Output Node name of the returned String. */
  private static String FORMATTED_NAME = "Dicom time";

  /** Output Node description of the returned String. */
  private static String FORMATTED_DESC = "Dicom time formatted string.";

  /**
   * Constructs a DicomTimeFormatter with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public DicomTimeFormatter(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a DicomTimeFormatter using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public DicomTimeFormatter(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Create the Input Node
      Class type = Integer.class;
      InputNode inputNode = new InputNode(HOUR_NAME, HOUR_DESC, type);
      _addInputNode(inputNode);

      // Create the Input Node
      type = Integer.class;
      inputNode = new InputNode(MINUTE_NAME, MINUTE_DESC, type);
      _addInputNode(inputNode);

      // Create the Input Node
      type = Integer.class;
      inputNode = new InputNode(SECOND_NAME, SECOND_DESC, type);
      _addInputNode(inputNode);

      // Create the Input Node
      type = Integer.class;
      inputNode = new InputNode(MILLISECOND_NAME, MILLISECOND_DESC, type);
      _addInputNode(inputNode);

      // Create the Output Node
      type = String.class;
      OutputNode outputNode = new OutputNode(FORMATTED_NAME, FORMATTED_DESC,
					     type);
      _addOutputNode(outputNode);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {
	//Get the Hour
	String hour = 
	  ((Integer)_getInputValue(HOUR_NAME, new Integer("0"))).toString();

	//Get the Minute
	String minute = 
	  ((Integer)_getInputValue(MINUTE_NAME, new Integer("0"))).toString();

	//Get the Time
	String second = 
	  ((Integer)_getInputValue(SECOND_NAME, new Integer("0"))).toString();

	//Get the Time
	String millisecond = 
	  ((Integer)_getInputValue(MILLISECOND_NAME, 
				   new Integer("0"))).toString();
	
	while(hour.length() < 2){
	  hour = "0" + hour;
	}
	hour = hour.substring(0,2);

	while(minute.length() < 2){
	  minute = "0" + minute;
	}
	minute = minute.substring(0,2);

	while(second.length() < 2){
	  second = "0" + second;
	}
	second = second.substring(0,2);

	while(millisecond.length() < 3){
	  millisecond = "0" + millisecond;
	}
	millisecond = millisecond.substring(0,3);
	
	_setOutputValue(FORMATTED_NAME, hour+minute+second+"."+millisecond);

      }catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }	

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      return new ProcessorProperty[0];
    }
}



      
