/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.vann;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.RangeRestriction;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.image.DataBufferInt;
import java.awt.image.DataBufferUShort;
import java.awt.image.WritableRaster;
import java.lang.reflect.Array;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Vector; 
import javax.imageio.ImageTypeSpecifier;

/**
 * Processor that converts a raw buffered image into an array of buffered 
 * images.
 *
 * @author Bryan C. Vann
 * 
 * @version 30 October 2003
 */
public class RawImageConverter extends Processor
{   
  /** Description of this Processor. */
  private static String DESC = "Converts date elements to dicom format.";

  /** Base Input Node name Buffered image. */
  private static String BUFFERED_IMAGE_NAME = "Buffered image"; 

  /** Input Node Buffered image description. */
  private static String BUFFERED_IMAGE_DESC = "Input buffered image";

  /** Base Input Node Skip Bytes name. */
  private static String SKIP_BYTES_NAME = "Skip Bytes"; 

  /** Input Node Skip Bytes description. */
  private static String SKIP_BYTES_DESC = "Bytes in header to skip";

  /** Base Input Node width name. */
  private static String WIDTH_NAME = "Width"; 

  /** Input Node width description. */
  private static String WIDTH_DESC = "Width of the image";

  /** Base Input Node Height name. */
  private static String HEIGHT_NAME = "Height"; 

  /** Input Node Height description. */
  private static String HEIGHT_DESC = "Height of the image";

  /** Base Input Node Swapped name. */
  private static String SWAPPED_NAME = "Swapped"; 

  /** Input Node Swapped description. */
  private static String SWAPPED_DESC = "Boolean to indicate byte swapping";

  /** Base Input Node Color type name. */
  private static String COLOR_TYPE_NAME = "Color type"; 

  /** Input Node Color type description. */
  private static String COLOR_TYPE_DESC = "Color type of the image";

  /** Output Node name of the returned array. */
  private static String ARRAY_NAME = "BufferedImage array";

  /** Output Node description of the returned String. */
  private static String ARRAY_DESC = "Array of buffered images.";

  /** 8 bit grayscale Input Node value. */
  public static String BYTE_GRAYSCALE = "8-bit grayscale";

  /** 16 bit grayscale Input Node value. */
  public static String USHORT_GRAYSCALE = "16-bit grayscale";

  /** RGB interleaved Input Node value. */
  public static String RGB_INTERLEAVED = "RGB interleaved";

  /** RGB banded Input Node value. */
  public static String RGB_BANDED = "RGB banded";

  /**
   * Constructs a RawImageConverter with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public RawImageConverter(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a RawImageConverter using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public RawImageConverter(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Create buffered image Input Node
      Class type = BufferedImage.class;
      InputNode inputNode = 
	new InputNode(BUFFERED_IMAGE_NAME, BUFFERED_IMAGE_DESC, type);
      _addInputNode(inputNode);

      // Create skip bytes Input Node
      inputNode = 
	new InputNode(SKIP_BYTES_NAME, SKIP_BYTES_DESC, Integer.class);
      _addInputNode(inputNode);

      // Create width Input Node
      inputNode = 
	new InputNode(WIDTH_NAME, WIDTH_DESC, Integer.class);
      _addInputNode(inputNode);

      // Create height Input Node
      inputNode = 
	new InputNode(HEIGHT_NAME, HEIGHT_DESC, Integer.class);
      _addInputNode(inputNode);

      // Create swapped Input Node
      inputNode = 
	new InputNode(SWAPPED_NAME, SWAPPED_DESC, Boolean.class);
      _addInputNode(inputNode);

      // Create color type Input Node
      inputNode = 
	new InputNode(COLOR_TYPE_NAME, COLOR_TYPE_DESC, String.class);
 
      // Create the Vector of supported color types.
      Vector ops = new Vector();
      ops.add(BYTE_GRAYSCALE);
      ops.add(USHORT_GRAYSCALE);
      ops.add(RGB_INTERLEAVED);
      ops.add(RGB_BANDED);

      String REST_NAME = "Allowable color types";
      String REST_DESC = "Allowable color types for this processor";
      Restriction restriction = new SetRestriction(REST_NAME, REST_DESC, ops);
      inputNode.setRestriction(restriction);
      _addInputNode(inputNode);

      // Create the Output Node
      type = Array.newInstance(BufferedImage.class, 0).getClass();
      OutputNode outputNode = 
	new OutputNode(ARRAY_NAME, ARRAY_DESC, type);
      _addOutputNode(outputNode);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {      
	//Get the buffered image
	BufferedImage image = 
	  (BufferedImage)_getInputValue(BUFFERED_IMAGE_NAME, null);

	//Get the number bytes to skip for the header
	int skipBytes = ((Integer)_getInputValue(SKIP_BYTES_NAME, 
						 new Integer("0"))).intValue();

	//Get the height
	int height = ((Integer)_getInputValue(HEIGHT_NAME, 
					    new Integer("0"))).intValue();

	//Get the width
	int width = ((Integer)_getInputValue(WIDTH_NAME, 
					    new Integer("0"))).intValue();

	//Get the byte swapping
	boolean swapped = ((Boolean)_getInputValue(SWAPPED_NAME, 
					   new Boolean(false))).booleanValue();
      
	//Get the color type
	String colorType = 
	  (String)_getInputValue(COLOR_TYPE_NAME, "8-bit grayscale");
	
	byte[] data = 
	  ((DataBufferByte)(image.getRaster().getDataBuffer())).getData();
	int pos = skipBytes;

	int numBytes;
	//int type;
	ImageTypeSpecifier type;
	if(colorType.equals(BYTE_GRAYSCALE)){
	  //type = BufferedImage.TYPE_BYTE_GRAY;
	  type = Utilities.getGrayImageType(width, height, 8);
	  numBytes = 1;
	}else if(colorType.equals(USHORT_GRAYSCALE)){
	  //type = BufferedImage.TYPE_USHORT_GRAY;
	  type = Utilities.getGrayImageType(width, height, 16);
	  numBytes = 2;
	}else{
	  //type = BufferedImage.TYPE_INT_RGB;
	  type = Utilities.getRgbImageType(width, height);
	  numBytes = 3;
	}
	int imageSize = height * width * numBytes;
	int numImages = (int)((data.length - pos)/imageSize);
	BufferedImage[] images = new BufferedImage[numImages];

	// Create each buffered Image
	for(int i = 0 ; i < numImages ; i++){
	  byte[] temp = new byte[imageSize];
	  System.arraycopy(data, pos, temp, 0, imageSize);

	  // Get the destination writable raster
	  BufferedImage bufferedImage = type.createBufferedImage(width,height);
	  WritableRaster dstRaster = bufferedImage.getRaster();

	  if(colorType.equals(BYTE_GRAYSCALE)){
	    images[i] = Utilities.getGrayByteImage(temp, dstRaster);
	  }else if(colorType.equals(USHORT_GRAYSCALE)){
	    images[i] = Utilities.getGrayUshortImage(temp, dstRaster, 16, 
						    false, swapped, 1, 0);
	  }else if(colorType.equals(RGB_INTERLEAVED)){
	    images[i] = Utilities.getRgbImage(temp, dstRaster, true);
	  }else{
	    images[i] = Utilities.getRgbImage(temp, dstRaster, false);
	  }
	  pos += imageSize;
	}
	_setOutputValue(ARRAY_NAME, images);
      }catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }	

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      return new ProcessorProperty[0];
    }
}


